# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing all life cycle management capabilities."""
try:
    import SherlockLifeCycleService_pb2
    import SherlockLifeCycleService_pb2_grpc
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockLifeCycleService_pb2
    from ansys.api.sherlock.v0 import SherlockLifeCycleService_pb2_grpc

import grpc

from ansys.sherlock.core import LOG
from ansys.sherlock.core.errors import (
    SherlockAddHarmonicEventError,
    SherlockAddHarmonicVibeProfilesError,
    SherlockAddRandomVibeEventError,
    SherlockAddRandomVibeProfilesError,
    SherlockAddShockEventError,
    SherlockAddShockProfilesError,
    SherlockAddThermalEventError,
    SherlockAddThermalProfilesError,
    SherlockCreateLifePhaseError,
    SherlockInvalidHarmonicProfileEntriesError,
    SherlockInvalidLoadDirectionError,
    SherlockInvalidOrientationError,
    SherlockInvalidRandomVibeProfileEntriesError,
    SherlockInvalidShockProfileEntriesError,
    SherlockInvalidThermalProfileEntriesError,
    SherlockLoadHarmonicProfileError,
    SherlockLoadRandomVibeProfileError,
    SherlockLoadShockProfileDatasetError,
    SherlockLoadShockProfilePulsesError,
    SherlockLoadThermalProfileError,
    SherlockNoGrpcConnectionException,
)
from ansys.sherlock.core.grpc_stub import GrpcStub
from ansys.sherlock.core.utils.version_check import require_version


class Lifecycle(GrpcStub):
    """Contains all life cycle management capabilities."""

    def __init__(self, channel: grpc.Channel, server_version: int):
        """Initialize a gRPC stub for the Sherlock Life Cycle service."""
        super().__init__(channel, server_version)
        self.stub = SherlockLifeCycleService_pb2_grpc.SherlockLifeCycleServiceStub(channel)
        self.CYCLE_TYPE_LIST = None
        self.RV_PROFILE_TYPE_LIST = None
        self.HARMONIC_PROFILE_TYPE_LIST = None
        self.AMPL_UNIT_LIST = None
        self.CYCLE_STATE_LIST = None
        self.LOAD_UNIT_LIST = None
        self.SHOCK_SHAPE_LIST = None
        self.STEP_TYPE_LIST = ["RAMP", "HOLD"]

    def _init_cycle_types(self):
        """Initialize the list for cycle types.

        Available Since: 2021R1
        """
        if self._is_connection_up():
            cycle_type_request = SherlockLifeCycleService_pb2.ListLCTypesRequest()
            cycle_type_response = self.stub.listLifeCycleTypes(cycle_type_request)
            if cycle_type_response.returnCode.value == 0:
                self.CYCLE_TYPE_LIST = cycle_type_response.types

    def _init_rv_profile_types(self):
        """Initialize the list for RV profile types.

        Available Since: 2023R1
        """
        if self._is_connection_up():
            rv_profile_request = SherlockLifeCycleService_pb2.ListRandomVibeProfileTypesRequest()
            rv_profile_response = self.stub.listRandomVibeProfileTypes(rv_profile_request)
            if rv_profile_response.returnCode.value == 0:
                self.RV_PROFILE_TYPE_LIST = rv_profile_response.types

    def _init_harmonic_profile_types(self):
        """Initialize the list for harmonic profile types.

        Available Since: 2021R1
        """
        if self._is_connection_up():
            harmonic_profile_request = (
                SherlockLifeCycleService_pb2.ListHarmonicProfileTypesRequest()
            )
            harmonic_profile_response = self.stub.listHarmonicProfileTypes(harmonic_profile_request)
            if harmonic_profile_response.returnCode.value == 0:
                self.HARMONIC_PROFILE_TYPE_LIST = harmonic_profile_response.types

    def _init_ampl_units(self):
        """Initialize the list for amplitude units.

        Available Since: 2021R1
        """
        if self._is_connection_up():
            ampl_unit_request = SherlockLifeCycleService_pb2.ListAmplUnitsRequest()
            ampl_type_response = self.stub.listAmplUnits(ampl_unit_request)
            if ampl_type_response.returnCode.value == 0:
                self.AMPL_UNIT_LIST = ampl_type_response.amplUnits

    def _init_cycle_states(self):
        """Initialize the list for cycle states.

        Available Since: 2021R1
        """
        if self._is_connection_up():
            cycle_state_request = SherlockLifeCycleService_pb2.ListLCStatesRequest()
            cycle_state_response = self.stub.listLifeCycleStates(cycle_state_request)
            if cycle_state_response.returnCode.value == 0:
                self.CYCLE_STATE_LIST = cycle_state_response.states

    def _init_load_units(self):
        """Initialize the list for load units.

        Available Since: 2021R1
        """
        if self._is_connection_up():
            load_unit_request = SherlockLifeCycleService_pb2.ListShockLoadUnitsRequest()
            load_unit_response = self.stub.listShockLoadUnits(load_unit_request)
            if load_unit_response.returnCode.value == 0:
                self.LOAD_UNIT_LIST = load_unit_response.units

    def _init_shock_shapes(self):
        """Initialize the list for shock shapes.

        Available Since: 2021R1
        """
        if self._is_connection_up():
            shock_shape_request = SherlockLifeCycleService_pb2.ListShockPulsesRequest()
            shock_shape_response = self.stub.listShockPulses(shock_shape_request)
            if shock_shape_response.returnCode.value == 0:
                self.SHOCK_SHAPE_LIST = shock_shape_response.shockPulse

    @staticmethod
    def _check_load_direction_validity(load_direction: str):
        """Check that the input string is a valid load."""
        directions = load_direction.split(",")

        if len(directions) != 3:
            raise SherlockInvalidLoadDirectionError("Number of direction coordinates is invalid.")

        try:
            nonzero = 0
            for direction in directions:
                if float(direction) != 0:
                    nonzero += 1

            if nonzero == 0:
                raise SherlockInvalidLoadDirectionError(
                    "At least one direction coordinate must be non-zero."
                )
            return
        except TypeError:
            raise SherlockInvalidLoadDirectionError("Direction coordinates are invalid.")

    @staticmethod
    def _check_orientation_validity(orientations: str):
        """Check input string if it is a valid orientation."""
        orientation = orientations.split(",")

        if len(orientation) != 2:
            raise SherlockInvalidOrientationError("Number of spherical coordinates is invalid.")

        try:
            float(orientation[0])
        except:
            raise SherlockInvalidOrientationError("Azimuth value is invalid.")

        try:
            float(orientation[1])
            return
        except:
            raise SherlockInvalidOrientationError("Elevation value is invalid.")

    @staticmethod
    def _check_random_vibe_profile_entries_validity(profile_entries: list):
        """Check input list to see if all elements are valid for random vibe entries."""
        if not isinstance(profile_entries, list):
            raise SherlockInvalidRandomVibeProfileEntriesError("Entries argument is invalid.")

        i = 0
        try:
            for i, entry in enumerate(profile_entries):
                if len(entry) != 2:
                    raise SherlockInvalidRandomVibeProfileEntriesError(
                        f"Invalid entry {i}: Number of elements is wrong"
                    )
                if entry[0] <= 0:
                    raise SherlockInvalidRandomVibeProfileEntriesError(
                        f"Invalid entry {i}: Frequencies must be greater than 0"
                    )
                if entry[1] <= 0:
                    raise SherlockInvalidRandomVibeProfileEntriesError(
                        f"Invalid entry {i}: Amplitudes must be greater than 0"
                    )
        except TypeError:
            raise SherlockInvalidRandomVibeProfileEntriesError(
                f"Invalid entry {i}: Frequency or amplitude is invalid"
            )

    def _check_thermal_profile_entries_validity(self, profile_entries: list):
        """Check input list to see if all elements are valid for thermal entries."""
        if not isinstance(profile_entries, list):
            raise SherlockAddThermalProfilesError("Entries argument is invalid.")

        i = 0
        try:
            for i, entry in enumerate(profile_entries):
                if len(entry) != 4:
                    raise SherlockInvalidThermalProfileEntriesError(
                        f"Invalid entry {i}: Number of elements is wrong"
                    )
                if not isinstance(entry[0], str):
                    raise SherlockInvalidThermalProfileEntriesError(
                        f"Invalid entry {i}: Step name is invalid"
                    )
                if entry[1] not in self.STEP_TYPE_LIST:
                    raise SherlockInvalidThermalProfileEntriesError(
                        f"Invalid entry {i}: Step type is invalid"
                    )
                if entry[2] <= 0:
                    raise SherlockInvalidThermalProfileEntriesError(
                        f"Invalid entry {i}: Time must be greater than 0"
                    )
                if not isinstance(entry[3], (int, float)):
                    raise SherlockInvalidThermalProfileEntriesError(
                        f"Invalid entry {i}: Temperature is invalid"
                    )
        except TypeError:
            raise SherlockInvalidThermalProfileEntriesError(f"Invalid entry {i}: Time is invalid")

    @staticmethod
    def _check_harmonic_profile_entries_validity(profile_entries: list):
        """Check input list if all elements are valid for harmonic entries."""
        if not isinstance(profile_entries, list):
            raise SherlockInvalidHarmonicProfileEntriesError(message="Entries argument is invalid.")

        i = 0
        try:
            for i, entry in enumerate(profile_entries):
                if len(entry) != 2:
                    raise SherlockInvalidHarmonicProfileEntriesError(
                        message=f"Invalid entry {i}: Number of elements is wrong"
                    )
                if entry[0] <= 0:
                    raise SherlockInvalidHarmonicProfileEntriesError(
                        message=f"Invalid entry {i}: Frequencies must be greater than 0"
                    )
                if entry[1] <= 0:
                    raise SherlockInvalidHarmonicProfileEntriesError(
                        message=f"Invalid entry {i}: Load must be greater than 0"
                    )
        except TypeError:
            raise SherlockInvalidHarmonicProfileEntriesError(
                message=f"Invalid entry {i}: Frequency or load is invalid"
            )

    def _check_shock_profile_entries_validity(self, profile_entries: list):
        """Check input list to see if all elements are valid for shock entries."""
        if not isinstance(profile_entries, list):
            raise SherlockInvalidShockProfileEntriesError(message="Entries argument is invalid.")

        i = 0
        try:
            for i, entry in enumerate(profile_entries):
                if len(entry) != 4:
                    raise SherlockInvalidShockProfileEntriesError(
                        message=f"Invalid entry {i}: Number of elements is wrong"
                    )
                if not isinstance(entry[0], str):
                    raise SherlockInvalidShockProfileEntriesError(
                        message=f"Invalid entry {i}: Shape name is invalid"
                    )
                if (self.SHOCK_SHAPE_LIST is not None) and (entry[0] not in self.SHOCK_SHAPE_LIST):
                    raise SherlockInvalidShockProfileEntriesError(
                        message=f"Invalid entry {i}: Shape type is invalid"
                    )
                if entry[1] <= 0:
                    raise SherlockInvalidShockProfileEntriesError(
                        message=f"Invalid entry {i}: Load must be greater than 0"
                    )
                if entry[2] <= 0:
                    raise SherlockInvalidShockProfileEntriesError(
                        message=f"Invalid entry {i}: Frequency must be greater than 0"
                    )
                if entry[3] < 0:
                    raise SherlockInvalidShockProfileEntriesError(
                        message=f"Invalid entry {i}: Decay must be non-negative"
                    )
            return
        except TypeError:
            raise SherlockInvalidShockProfileEntriesError(
                message=f"Invalid entry {i}: Load, frequency, or decay is invalid"
            )

    @require_version()
    def create_life_phase(
        self,
        project: str,
        phase_name: str,
        duration: float,
        duration_units: str,
        num_of_cycles: float,
        cycle_type: str,
        description: str = "",
    ):
        """Create a life phase.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        phase_name: str
            Name of the life phase.
        duration: float
            Event duration length.
        duration_units: str
            Units for the event duration length. Options are ``"ms"``,
            ``"sec"``, and ``"min"``.
        num_of_cycles: float
            Number of cycles for the life phase.
        cycle_type: str
            Cycle type. Options include ``"COUNT"``, ``"DUTY CYCLE"``,
            ``"PER YEAR"``, and ``"PER HOUR"``.
        description: str, optional
            Description of the life phase. The default is ``""``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
        )
        >>> sherlock.lifecycle.create_life_phase(
            "Test",
            "Example",
            1.5,
            "sec",
            4.0,
            "COUNT"
        )
        """
        if self.CYCLE_TYPE_LIST is None:
            self._init_cycle_types()

        try:
            if project == "":
                raise SherlockCreateLifePhaseError(message="Project name is invalid.")
            if phase_name == "":
                raise SherlockCreateLifePhaseError(message="Phase name is invalid.")
            if duration <= 0.0:
                raise SherlockCreateLifePhaseError(message="Duration must be greater than 0.")
            if (self.CYCLE_TYPE_LIST is not None) and (cycle_type not in self.CYCLE_TYPE_LIST):
                raise SherlockCreateLifePhaseError(message="Cycle type is invalid.")
            if num_of_cycles <= 0.0:
                raise SherlockCreateLifePhaseError(
                    message="Number of cycles must be greater than 0."
                )
        except SherlockCreateLifePhaseError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockLifeCycleService_pb2.CreateLifePhaseRequest(
            project=project,
            phaseName=phase_name,
            description=description,
            duration=duration,
            durationUnits=duration_units,
            numOfCycles=num_of_cycles,
            cycleType=cycle_type,
        )

        response = self.stub.createLifePhase(request)

        return_code = response.returnCode

        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockCreateLifePhaseError(error_array=response.errors)

                raise SherlockCreateLifePhaseError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockCreateLifePhaseError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def add_random_vibe_event(
        self,
        project: str,
        phase_name: str,
        event_name: str,
        duration: float,
        duration_units: str,
        num_of_cycles: float,
        cycle_type: str,
        orientation: str,
        profile_type: str,
        load_direction: str,
        description: str = "",
    ) -> int:
        """Add a random vibe event to a life cycle phase.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        phase_name: str
            Name of the life cycle phase to add the random vibe event to.
        event_name: str
            Name of the random vibe event.
        duration: float
            Event duration length.
        duration_units: str
            Event duration units. Options are ``"ms"``, ``"sec"``, ``"min"``,
            ``"hr"``, ``"day"``, and ``"year"``.
        num_of_cycles: float
            Number of cycles for the random vibe event.
        cycle_type: str
            Cycle type. Options are ``"COUNT"``, ``"DUTY_CYCLE"``, ``"PER_YEAR"``,
            ``"PER_DAY"``, ``"PER_HOUR"``, ``"PER_MIN"``, and ``"PER_SEC"``.
        orientation: str
            PCB orientation in the format of ``"azimuth, elevation"``. For example,
            ``"30,15"``.
        profile_type: str
            Random load profile type. The only option is ``"Uniaxial"``.
        load_direction: str
            Load direction in the format of ``"x,y,z"``. For example, ``"0,0,1"``.
        description: str, optional
            Description of the random vibe event. The default is ``""``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
        )
        >>> sherlock.lifecycle.create_life_phase(
            "Test",
            "Example",
            1.5,
            "sec",
            4.0,
            "COUNT",
        )
        >>> sherlock.lifecycle.add_random_vibe_event(
            "Test",
            "Example",
            "Event1",
            1.5,
            "sec",
            4.0,
            "PER MIN",
            "45,45",
            "Uniaxial",
            "2,4,5"
        )
        """
        if self.CYCLE_TYPE_LIST is None:
            self._init_cycle_types()
        if self.RV_PROFILE_TYPE_LIST is None:
            self._init_rv_profile_types()

        try:
            if project == "":
                raise SherlockAddRandomVibeEventError(message="Project name is invalid.")
            if phase_name == "":
                raise SherlockAddRandomVibeEventError(message="Phase name is invalid.")
            if event_name == "":
                raise SherlockAddRandomVibeEventError(message="Event name is invalid.")
            if duration <= 0.0:
                raise SherlockAddRandomVibeEventError(message="Duration must be greater than 0.")
            if (self.CYCLE_TYPE_LIST is not None) and (cycle_type not in self.CYCLE_TYPE_LIST):
                raise SherlockAddRandomVibeEventError(message="Cycle type is invalid.")
            if num_of_cycles <= 0.0:
                raise SherlockAddRandomVibeEventError(
                    message="Number of cycles must be greater than 0."
                )
        except SherlockAddRandomVibeEventError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        try:
            self._check_load_direction_validity(load_direction)
            if (self.RV_PROFILE_TYPE_LIST is not None) and (
                profile_type not in self.RV_PROFILE_TYPE_LIST
            ):
                raise SherlockAddRandomVibeEventError(message="Invalid profile type.")
            self._check_orientation_validity(orientation)
        except (SherlockInvalidLoadDirectionError, SherlockInvalidOrientationError) as e:
            LOG.error(f"Add random vibe event error: {str(e)}")
            raise SherlockAddRandomVibeEventError(message=str(e))
        except SherlockAddRandomVibeEventError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockLifeCycleService_pb2.AddRandomVibeEventRequest(
            project=project,
            phaseName=phase_name,
            eventName=event_name,
            description=description,
            duration=duration,
            durationUnits=duration_units,
            numOfCycles=num_of_cycles,
            cycleType=cycle_type,
            orientation=orientation,
            profileType=profile_type,
            loadDirection=load_direction,
        )

        response = self.stub.addRandomVibeEvent(request)
        return_code = response.returnCode
        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockAddRandomVibeEventError(error_array=response.errors)

                raise SherlockAddRandomVibeEventError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockAddRandomVibeEventError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def add_random_vibe_profiles(
        self,
        project: str,
        random_vibe_profiles: list[tuple[str, str, str, str, str, list[tuple[float, float]]]],
    ):
        """Add random vibe profiles to a life cycle phase.

        Available Since: 2023R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        random_vibe_profiles: list[tuple[str, str, str, str, str, list[tuple[float, float]]]]
            Random vibe profiles consisting of these properties:

            - phase_name: str
                Name of the life cycle phase to add the random vibe profile to.
            - event_name: str
                Name of the random vibe event.
            - profile_name: str
                Name of the random vibe profile.
            - freq_units: str
                Frequency units. Options are ``"HZ"``, ``"KHZ"``, ``"MHZ"``, and ``"GHZ"``.
            - ampl_units: str
                Amplitude units. Options are ``"G2/Hz"``, ``"m2/s4/Hz"``, ``"mm2/s4/Hz"``, \
                ``"in2/s4/Hz"``, and ``"ft2/s4/Hz"``.
            - random_vibe_profile_entries: list[tuple[float, float]]
                Random vibe profile entries consisting of these properties:

                - frequency: float
                    Frequency of the profile entry expressed in frequency units.
                - amplitude: float
                    Amplitude of the profile entry expressed in amplitude units.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
        )
        >>> sherlock.lifecycle.create_life_phase(
            "Test",
            "Example",
            1.5,
            "sec",
            4.0,
            "COUNT",
        )
        >>> sherlock.lifecycle.add_random_vibe_event(
            "Test",
            "Example",
            "Event1",
            1.5,
            "sec",
            4.0,
            "PER MIN",
            "45,45",
            "Uniaxial",
            "2,4,5",
        )
        >>> sherlock.lifecycle.add_random_vibe_profiles(
            "Test",
             [(
                "Example",
                "Event1",
                "Profile1",
                "HZ",
                "G2/Hz",
                [(4,8), (5, 50)],
            )]
        )
        """
        if self.AMPL_UNIT_LIST is None:
            self._init_ampl_units()

        try:
            if project == "":
                raise SherlockAddRandomVibeProfilesError(message="Project name is invalid.")

            if len(random_vibe_profiles) == 0:
                raise SherlockAddRandomVibeProfilesError(
                    message="Random vibe profiles are " f"missing."
                )

            for i, profile_entry in enumerate(random_vibe_profiles):
                if len(profile_entry) != 6:
                    raise SherlockAddRandomVibeProfilesError(
                        f"Number of elements ({str(len(profile_entry))}) is wrong for "
                        f"random vibe profile {i}."
                    )
                elif not isinstance(profile_entry[0], str) or profile_entry[0] == "":
                    raise SherlockAddRandomVibeProfilesError(
                        f"Phase name is invalid for random vibe profile {i}."
                    )
                elif not isinstance(profile_entry[1], str) or profile_entry[1] == "":
                    raise SherlockAddRandomVibeProfilesError(
                        f"Event name is invalid for random vibe profile {i}."
                    )
                elif not isinstance(profile_entry[2], str) or profile_entry[2] == "":
                    raise SherlockAddRandomVibeProfilesError(
                        f"Profile name is invalid for random vibe profile {i}."
                    )
                elif not isinstance(profile_entry[4], str) or (
                    (self.AMPL_UNIT_LIST is not None)
                    and (profile_entry[4] not in self.AMPL_UNIT_LIST)
                ):
                    raise SherlockAddRandomVibeProfilesError(
                        f"Amplitude type {profile_entry[4]} is invalid for random vibe profile {i}."
                    )

                try:
                    self._check_random_vibe_profile_entries_validity(profile_entry[5])
                except SherlockInvalidRandomVibeProfileEntriesError as e:
                    raise SherlockAddRandomVibeProfilesError(
                        f"{str(e)} for random vibe profile {i}."
                    )

        except SherlockAddRandomVibeProfilesError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockLifeCycleService_pb2.AddRandomVibeProfilesRequest(project=project)

        """Add random vibe profiles to the request."""
        for r in random_vibe_profiles:
            profile = request.randomVibeProfiles.add()
            profile.phaseName = r[0]
            profile.eventName = r[1]
            profile.profileName = r[2]
            profile.freqUnits = r[3]
            profile.amplUnits = r[4]

            """Add random vibe entries to the request."""
            for e in r[5]:
                entry = profile.randomVibeProfileEntries.add()
                entry.freq = e[0]
                entry.ampl = e[1]

        response = self.stub.addRandomVibeProfiles(request)

        return_code = response.returnCode

        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockAddRandomVibeProfilesError(error_array=response.errors)
                else:
                    raise SherlockAddRandomVibeProfilesError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockAddRandomVibeProfilesError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def add_thermal_event(
        self,
        project: str,
        phase_name: str,
        event_name: str,
        num_of_cycles: float,
        cycle_type: str,
        cycle_state: str,
        description: str = "",
    ) -> int:
        """Add a thermal event to a life cycle phase.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        phase_name: str
            Name of the life cycle phase to add the thermal event to.
        event_name: str
            Name of the thermal event.
        num_of_cycles: float
            Number of cycles for the thermal event.
        cycle_type: str
            Cycle type. Options are ``"COUNT"``, ``"DUTY_CYCLE"``, ``"PER_YEAR"``,
            ``"PER_DAY"``, ``"PER_HOUR"``, ``"PER_MIN"``, and ``"PER_SEC"``.
        cycle_state: str
            Life cycle state. Options are ``"OPERATING"`` and ``"STORAGE"``.
        description: str, optional
            Description of the thermal event. The default is ``""``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
        )
        >>> sherlock.lifecycle.create_life_phase(
            "Test",
            "Example",
            1.5,
            "year",
            4.0,
            "COUNT",
        )
        >>> sherlock.lifecycle.add_thermal_event(
            "Test",
            "Example",
            "Event1",
            4.0,
            "PER YEAR",
            "STORAGE"
        )
        """
        if self.CYCLE_TYPE_LIST is None:
            self._init_cycle_types()
        if self.CYCLE_STATE_LIST is None:
            self._init_cycle_states()

        try:
            if project == "":
                raise SherlockAddThermalEventError(message="Project name is invalid.")
            if phase_name == "":
                raise SherlockAddThermalEventError(message="Phase name is invalid.")
            if event_name == "":
                raise SherlockAddThermalEventError(message="Event name is invalid.")
            if (self.CYCLE_TYPE_LIST is not None) and (cycle_type not in self.CYCLE_TYPE_LIST):
                raise SherlockAddThermalEventError(message="Cycle type is invalid.")
            if num_of_cycles <= 0.0:
                raise SherlockAddThermalEventError(
                    message="Number of cycles must be greater than 0."
                )
            if (self.CYCLE_STATE_LIST is not None) and (cycle_state not in self.CYCLE_STATE_LIST):
                raise SherlockAddThermalEventError(message="Cycle state is invalid.")
        except SherlockAddThermalEventError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        request = SherlockLifeCycleService_pb2.AddThermalEventRequest(
            project=project,
            phaseName=phase_name,
            eventName=event_name,
            description=description,
            numOfCycles=num_of_cycles,
            cycleType=cycle_type,
            cycleState=cycle_state,
        )

        response = self.stub.addThermalEvent(request)

        return_code = response.returnCode

        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockAddThermalEventError(error_array=response.errors)

                raise SherlockAddThermalEventError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockAddThermalEventError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def add_thermal_profiles(
        self,
        project: str,
        thermal_profiles: list[tuple[str, str, str, str, str, list[tuple[str, str, float, float]]]],
    ) -> int:
        """Add thermal profiles to a life cycle phase.

        Available Since: 2023R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        thermal_profiles: list[tuple[str, str, str, str, str, list[tuple[str, str, float, float]]]]
            Thermal profiles consisting of these properties:

            - phase_name: str
                Name of the life cycle phase to add the thermal profile to.
            - event_name: str
                Name of the thermal event.
            - profile_name: str
                Name of the thermal profile.
            - time_units: str
                Time units. Options are ``"ms"``, ``"sec"``, ``"min"``, ``"hr"``,
                ``"day"``, and ``"year"``.
            - temp_units: str
                Temperature units. Options are ``"C"``, ``"F"``, and ``"K"``.
            - thermal_profile_entries: list[tuple[str, str, float, float]]
                Thermal profile entries consisting of these properties:

                - step: str
                    Name of the thermal step.
                - type: str
                    Type of the thermal step. Options are ``"HOLD"`` and ``"RAMP"``.
                - time: float
                    Duration of the thermal step expressed in time units.
                - temperature: float
                    Temperature of the step expressed in temperature units.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
        )
        >>> sherlock.lifecycle.create_life_phase(
            "Test",
            "Example",
            1.5,
            "year",
            4.0,
            "COUNT",
        )
        >>> sherlock.lifecycle.add_thermal_event(
            "Test",
            "Example",
            "Event1",
            4.0,
            "PER YEAR",
            "STORAGE",
        )
        >>> sherlock.lifecycle.add_thermal_profiles(
            "Test",
            [(
                "Example",
                "Event1",
                "Profile1",
                "sec",
                "F",
                [
                    ("Steady1", "HOLD", 40, 40),
                    ("Steady", "HOLD", 20, 20),
                    ("Back", "RAMP", 20, 40),
                ],
            )]
        )
        """
        try:
            if project == "":
                raise SherlockAddThermalProfilesError(message="Project name is invalid.")

            if len(thermal_profiles) == 0:
                raise SherlockAddThermalProfilesError(message="Thermal profiles are missing.")

            for i, profile_entry in enumerate(thermal_profiles):
                if len(profile_entry) != 6:
                    raise SherlockAddThermalProfilesError(
                        f"Number of elements ({str(len(profile_entry))}) is wrong for "
                        f"thermal profile {i}."
                    )
                elif not isinstance(profile_entry[0], str) or profile_entry[0] == "":
                    raise SherlockAddThermalProfilesError(
                        f"Phase name is invalid for thermal profile {i}."
                    )
                elif not isinstance(profile_entry[1], str) or profile_entry[1] == "":
                    raise SherlockAddThermalProfilesError(
                        f"Event name is invalid for thermal profile {i}."
                    )
                elif not isinstance(profile_entry[2], str) or profile_entry[2] == "":
                    raise SherlockAddThermalProfilesError(
                        f"Profile name is invalid for thermal profile {i}."
                    )

                try:
                    self._check_thermal_profile_entries_validity(profile_entry[5])
                except (
                    SherlockAddThermalProfilesError,
                    SherlockInvalidThermalProfileEntriesError,
                ) as e:
                    raise SherlockAddThermalProfilesError(f"{str(e)} for thermal profile {i}.")

        except SherlockAddThermalProfilesError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockLifeCycleService_pb2.AddThermalProfilesRequest(project=project)

        """Add  thermal profiles to the request."""
        for t in thermal_profiles:
            profile = request.thermalProfiles.add()
            profile.phaseName = t[0]
            profile.eventName = t[1]
            profile.profileName = t[2]
            profile.timeUnits = t[3]
            profile.tempUnits = t[4]

            """Add thermal profile entries to the request."""
            for e in t[5]:
                entry = profile.thermalProfileEntries.add()
                entry.step = e[0]
                entry.type = e[1]
                entry.time = e[2]
                entry.temp = e[3]

        response = self.stub.addThermalProfiles(request)

        return_code = response.returnCode

        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockAddThermalProfilesError(error_array=response.errors)
                else:
                    raise SherlockAddThermalProfilesError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockAddThermalProfilesError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def add_harmonic_event(
        self,
        project: str,
        phase_name: str,
        event_name: str,
        duration: float,
        duration_units: str,
        num_of_cycles: float,
        cycle_type: str,
        sweep_rate: float,
        orientation: str,
        profile_type: str,
        load_direction: str,
        description: str = "",
    ) -> int:
        """Add a harmonic event to a life cycle phase.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        phase_name: str
            Name of the life cycle phase to add the harmonic event to.
        event_name: str
            Name of the harmonic event.
        duration: float
            Event duration length.
        duration_units: str
            Event duration units. Options are ``"ms"``, ``"sec"``, ``"min"``,
            ``"hr"``, ``"day"``, and ``"year"``.
        num_of_cycles: float
            Number of cycles for the harmonic event.
        cycle_type: str
            Cycle type. Options are ``"COUNT"``, ``"DUTY_CYCLE"``, ``"PER_YEAR"``,
            ``"PER_DAY"``, ``"PER_HOUR"``, ``"PER_MIN"``, and ``"PER_SEC"``.
        sweep_rate: float
            Sweep rate for the harmonic event.
        orientation: str
            PCB orientation in the format of ``"azimuth, elevation"``. For example,
            ``"30,15"``.
        profile_type: str
            Profile type of the harmonic load. Options are ``"Uniaxial"`` and ``"Triaxial"``.
        load_direction: str
            Load direction in the format of ``"x,y,z"``. For example, ``"0,0,1"``.
        description: str, optional
            Description of the harmonic event. The default is ``""``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test"
        )
        >>> sherlock.lifecycle.create_life_phase(
            "Test",
            "Example",
            1.5,
            "year",
            4.0,
            "COUNT",
        )
        >>> sherlock.lifecycle.add_harmonic_event(
            "Test",
            "Example",
            "Event1",
            1.5,
            "sec",
            4.0,
            "PER MIN",
            5,
            "45,45",
            "Uniaxial",
            "2,4,5"
        )
        """
        if self.CYCLE_TYPE_LIST is None:
            self._init_cycle_types()
        if self.HARMONIC_PROFILE_TYPE_LIST is None:
            self._init_harmonic_profile_types()

        try:
            if project == "":
                raise SherlockAddHarmonicEventError(message="Project name is invalid.")
            if phase_name == "":
                raise SherlockAddHarmonicEventError(message="Phase name is invalid.")
            if event_name == "":
                raise SherlockAddHarmonicEventError(message="Event name is invalid.")
            if duration <= 0.0:
                raise SherlockAddHarmonicEventError(message="Duration must be greater than 0.")
            if (self.CYCLE_TYPE_LIST is not None) and (cycle_type not in self.CYCLE_TYPE_LIST):
                raise SherlockAddHarmonicEventError(message="Cycle type is invalid.")
            if num_of_cycles <= 0.0:
                raise SherlockAddHarmonicEventError(
                    message="Number of cycles must be greater than 0."
                )
            if sweep_rate <= 0.0:
                raise SherlockAddHarmonicEventError(message="Sweep rate must be greater than 0.")
        except SherlockAddHarmonicEventError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        try:
            self._check_load_direction_validity(load_direction)
            self._check_orientation_validity(orientation)
            if (self.HARMONIC_PROFILE_TYPE_LIST is not None) and (
                profile_type not in self.HARMONIC_PROFILE_TYPE_LIST
            ):
                raise SherlockAddHarmonicEventError(message="Profile type is invalid.")
        except (SherlockInvalidLoadDirectionError, SherlockInvalidOrientationError) as e:
            LOG.error(f"Add harmonic event error: {str(e)}")
            raise SherlockAddHarmonicEventError(message=str(e))
        except SherlockAddHarmonicEventError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        request = SherlockLifeCycleService_pb2.AddHarmonicEventRequest(
            project=project,
            phaseName=phase_name,
            eventName=event_name,
            description=description,
            duration=duration,
            durationUnits=duration_units,
            numOfCycles=num_of_cycles,
            cycleType=cycle_type,
            sweepRate=sweep_rate,
            orientation=orientation,
            profileType=profile_type,
            loadDirection=load_direction,
        )

        response = self.stub.addHarmonicEvent(request)

        return_code = response.returnCode

        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockAddHarmonicEventError(error_array=response.errors)

                raise SherlockAddHarmonicEventError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockAddHarmonicEventError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def add_harmonic_vibe_profiles(
        self,
        project: str,
        harmonic_vibe_profiles: list[
            tuple[str, str, str, str, str, list[tuple[float, float, str]]]
        ],
    ) -> int:
        """Add harmonic vibe profiles to a life cycle phase.

        Available Since: 2023R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        harmonic_vibe_profiles: list
            Harmonic vibe profiles consisting of these properties:

            - phase_name: str
                Name of the life cycle phase to add this harmonic vibe profile to.
            - event_name: str
                Name of the event.
            - profile_name: str
                Name of the harmonic vibe profile.
            - freq_units: str
                Frequency units. Options are ``"HZ"``, ``"KHZ"``, ``"MHZ"``,
                and ``"GHZ"``.
            - load_units: str
                Load units. Options are ``"G"``, ```"m/s2"``, ``"mm/s2"``,
                ``"in/s2"``, and ``"ft/s2"``.
            - harmonic_profile_entries: list[tuple[float, float, str]]
                Harmonic profile entries consisting of these properties:

                - frequency: float
                    Frequency of the harmonic profile expressed in frequency units.
                - load: float
                    Load of the harmonic profile expressed in load units.
                - triaxial_axis: str
                    Axis that this profile should be assigned to if the harmonic
                    profile type is ``"Triaxial"``. Options are: ``"x"``, ``"y"``,
                    and ``"z"``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
        )
        >>> sherlock.lifecycle.create_life_phase(
            "Test",
            "Example",
            1.5,
            "sec",
            4.0,
            "COUNT",
        )
        >>> sherlock.lifecycle.add_harmonic_event(
            "Test",
            "Example",
            "Event1",
            1.5,
            "sec",
            4.0,
            "PER MIN",
            5,
            "45,45",
            "Uniaxial",
            "2,4,5",
        )
        >>> sherlock.lifecycle.add_harmonic_vibe_profiles(
            "Test",
            [(
                "Example",
                "Event1",
                "Profile1",
                "HZ",
                "G",
                [
                    (10, 1),
                    (1000, 1),
                ],
                "",
            )]
        )
        """
        if self.LOAD_UNIT_LIST is None:
            self._init_load_units()

        try:
            if project == "":
                raise SherlockAddHarmonicVibeProfilesError(message="Project name is invalid.")

            i = 0
            profile_entry = []
            for i, profile_entry in enumerate(harmonic_vibe_profiles):
                if len(profile_entry) != 7:
                    raise SherlockAddHarmonicVibeProfilesError(
                        f"Number of elements ({str(len(profile_entry))}) is wrong for "
                        f"harmonic vibe profile {i}."
                    )
                elif not isinstance(profile_entry[0], str) or profile_entry[0] == "":
                    raise SherlockAddHarmonicVibeProfilesError(
                        f"Phase name is invalid for harmonic vibe profile {i}."
                    )
                elif not isinstance(profile_entry[1], str) or profile_entry[1] == "":
                    raise SherlockAddHarmonicVibeProfilesError(
                        f"Event name is invalid for harmonic vibe profile {i}."
                    )
                elif not isinstance(profile_entry[2], str) or profile_entry[2] == "":
                    raise SherlockAddHarmonicVibeProfilesError(
                        f"Profile name is invalid for harmonic vibe profile {i}."
                    )
                elif not isinstance(profile_entry[4], str) or (
                    (self.LOAD_UNIT_LIST is not None)
                    and (profile_entry[4] not in self.LOAD_UNIT_LIST)
                ):
                    raise SherlockAddHarmonicVibeProfilesError(
                        f"Load units {profile_entry[4]} are invalid for harmonic vibe profile {i}."
                    )

            try:
                self._check_harmonic_profile_entries_validity(profile_entry[5])
            except SherlockInvalidHarmonicProfileEntriesError as e:
                raise SherlockAddHarmonicVibeProfilesError(
                    f"{str(e)} for harmonic vibe profile {i}."
                )

        except SherlockAddHarmonicVibeProfilesError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockLifeCycleService_pb2.AddHarmonicVibeProfilesRequest(project=project)

        """Add harmonic vibe profiles to the request."""
        for h in harmonic_vibe_profiles:
            profile = request.harmonicVibeProfiles.add()
            profile.phaseName = h[0]
            profile.eventName = h[1]
            profile.profileName = h[2]
            profile.freqUnits = h[3]
            profile.loadUnits = h[4]

            """Add entries to the harmonic profile request."""
            for e in h[5]:
                entry = profile.harmonicVibeProfileEntries.add()
                entry.freq = e[0]
                entry.load = e[1]

            profile.triaxialAxis = h[6]

        response = self.stub.addHarmonicVibeProfiles(request)

        return_code = response.returnCode

        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockAddHarmonicVibeProfilesError(error_array=response.errors)
                else:
                    raise SherlockAddHarmonicVibeProfilesError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockAddHarmonicVibeProfilesError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def add_shock_event(
        self,
        project: str,
        phase_name: str,
        event_name: str,
        duration: float,
        duration_units: str,
        num_of_cycles: float,
        cycle_type: str,
        orientation: str,
        load_direction: str,
        description: str = "",
    ) -> int:
        """Add a shock event to a life cycle phase.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        phase_name: str
            Name of the life cycle phase to add this shock event to.
        event_name: str
            Name of the shock event.
        duration: float
            Event duration length.
        duration_units: str
            Event duration units. Options are ``"ms"``, ``"sec"``, ``"min"``, ``"hr"``,
            ``"day"``, and ``"year"``.
        num_of_cycles: float
            Number of cycles for the shock event.
        cycle_type: str
            Cycle type. Options are ``"COUNT"``, ``"DUTY CYCLE"``,
            ``"PER YEAR"``, and ``"PER HOUR"``.
        orientation: str
            PCB orientation in the format of ``"azimuth, elevation"``. For example,
            ``"30,15"``.
        load_direction: str
            Load direction in the format of ``"x,y,z"``. For example, ``"0,0,1"``.
        description: str, optional
            Description of the shock event. The default is ``""``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
        )
        >>> sherlock.lifecycle.create_life_phase(
            "Test",
            "Example",
            1.5,
            "sec",
            4.0,
            "COUNT",
        )
        >>> sherlock.lifecycle.add_shock_event(
            "Test",
            "Example",
            "Event1",
            1.5,
            "sec",
            4.0,
            "PER MIN",
            "45,45",
            "2,4,5"
        )
        """
        if self.CYCLE_TYPE_LIST is None:
            self._init_cycle_types()

        try:
            if project == "":
                raise SherlockAddShockEventError(message="Project name is invalid.")
            if phase_name == "":
                raise SherlockAddShockEventError(message="Phase name is invalid.")
            if event_name == "":
                raise SherlockAddShockEventError(message="Event name is invalid.")
            if duration <= 0.0:
                raise SherlockAddShockEventError(message="Duration must be greater than 0.")
            if (self.CYCLE_TYPE_LIST is not None) and (cycle_type not in self.CYCLE_TYPE_LIST):
                raise SherlockAddShockEventError(message="Cycle type is invalid.")
            if num_of_cycles <= 0.0:
                raise SherlockAddShockEventError(message="Number of cycles must be greater than 0.")
        except SherlockAddShockEventError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        try:
            self._check_load_direction_validity(load_direction)
            self._check_orientation_validity(orientation)
        except (SherlockInvalidLoadDirectionError, SherlockInvalidOrientationError) as e:
            LOG.error(f"Add shock event error: {str(e)}")
            raise SherlockAddShockEventError(message=str(e))

        request = SherlockLifeCycleService_pb2.AddShockEventRequest(
            project=project,
            phaseName=phase_name,
            eventName=event_name,
            description=description,
            duration=duration,
            durationUnits=duration_units,
            numOfCycles=num_of_cycles,
            cycleType=cycle_type,
            orientation=orientation,
            loadDirection=load_direction,
        )

        response = self.stub.addShockEvent(request)

        return_code = response.returnCode

        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockAddShockEventError(error_array=response.errors)

                raise SherlockAddShockEventError(message=return_code.message)

            return return_code.value
        except SherlockAddShockEventError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def add_shock_profiles(
        self,
        project: str,
        shock_profiles: list[
            tuple[
                str,
                str,
                str,
                float,
                str,
                float,
                str,
                str,
                str,
                list[tuple[str, float, float, float]],
            ]
        ],
    ) -> int:
        """Add shock profiles to a life cycle phase.

        Available Since: 2023R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project
        shock_profiles: list
            Shock profiles consisting of these properties:

            - phase_name: str
                Name of the life cycle phase to add the shock profile to.
            - event_name: str
                Name of the shock event.
            - profile_name: str
                Name of the shock profile.
            - duration: float
                Pulse duration.
            - duration_units: str
                Pulse duration units. Options are ``"ms"``, ``"sec"``, ``"min"``, ``"hr"``,
                ``"day"``, and ``"year"``.
            - sample_rate: float
                Sample rate.
            - sample_rate_units: str
                Sample rate units. Options are ``"ms"``, ``"sec"``, ``"min"``, ``"hr"``,
                ``"day"``, and ``"year"``.
            - load_units: str
                Load units. Options are: ``"G"``, ``"m/s2"``, ``"mm/s2"``, ``"in/s2"``,
                and ``"ft/s2"``.
            - freq_units: str
                Frequency units. Options are ``"HZ"``, ``"KHZ"``, ``"MHZ"``, and ``"GHZ"``.
            - shock_profile_entries: list
                Shock profile entries consisting of these properties:

                - shape: str
                    Shape of the shock profile entry. Options are ``"FullSine"``,
                    ``"HalfSine"``, ``"Haversine"``, ``"Triangle"``, ``"Sawtooth"``,
                    ``"FullSquare"``, and ``"HalfSquare"``.
                - load: float
                    Load of the profile entry expressed in load units.
                - freq: float
                    Frequency of the profile entry expressed in frequency units.
                - decay: float
                    Decay value of the profile entry.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
        )
        >>> sherlock.lifecycle.create_life_phase(
            "Test",
            "Example",
            1.5,
            "sec",
            4.0,
            "COUNT",
        )
        >>> sherlock.lifecycle.add_shock_event(
            "Test",
            "Example",
            "Event1",
            1.5,
            "sec",
            4.0,
            "PER MIN",
            "45,45",
            "2,4,5",
        )
        >>> sherlock.lifecycle.add_shock_profiles(
            "Test",
            [(
                "Example",
                "Event1",
                "Profile1",
                10.0, "ms",
                0.1, "ms",
                "G",
                "HZ",
                [("HalfSine", 100.0, 100.0, 0)],
            )]
        )
        """
        if self.LOAD_UNIT_LIST is None:
            self._init_load_units()
        if self.SHOCK_SHAPE_LIST is None:
            self._init_shock_shapes()

        try:
            if project == "":
                raise SherlockAddShockProfilesError(message="Project name is invalid.")

            i = 0
            profile_entry = []
            for i, profile_entry in enumerate(shock_profiles):
                if len(profile_entry) != 10:
                    raise SherlockAddShockProfilesError(
                        f"Number of elements ({str(len(profile_entry))}) is wrong for shock "
                        f"profile {i}."
                    )
                elif not isinstance(profile_entry[0], str) or profile_entry[0] == "":
                    raise SherlockAddShockProfilesError(
                        f"Phase name is invalid for shock profile {i}."
                    )
                elif not isinstance(profile_entry[1], str) or profile_entry[1] == "":
                    raise SherlockAddShockProfilesError(
                        f"Event name is invalid for shock profile {i}."
                    )
                elif not isinstance(profile_entry[2], str) or profile_entry[2] == "":
                    raise SherlockAddShockProfilesError(
                        f"Profile name is invalid for shock profile {i}."
                    )
                elif profile_entry[3] <= 0:
                    raise SherlockAddShockProfilesError(
                        f"Duration must be greater than 0 for shock profile {i}."
                    )
                elif not isinstance(profile_entry[4], str):
                    raise SherlockAddShockProfilesError(
                        f"Duration units {profile_entry[4]} are invalid for shock profile {i}."
                    )
                elif profile_entry[5] <= 0:
                    raise SherlockAddShockProfilesError(
                        f"Sample rate must be greater than 0 for shock profile {i}."
                    )
                elif not isinstance(profile_entry[6], str):
                    raise SherlockAddShockProfilesError(
                        f"Sample rate unit {profile_entry[6]} are invalid for shock profile {i}."
                    )
                elif not isinstance(profile_entry[7], str) or (
                    (self.LOAD_UNIT_LIST is not None)
                    and (profile_entry[7] not in self.LOAD_UNIT_LIST)
                ):
                    raise SherlockAddShockProfilesError(
                        f"Load units {profile_entry[7]} are invalid for shock profile {i}."
                    )

            try:
                self._check_shock_profile_entries_validity(profile_entry[9])
            except SherlockInvalidShockProfileEntriesError as e:
                raise SherlockAddShockProfilesError(f"{str(e)} for shock profile {i}.")

        except SherlockAddShockProfilesError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockLifeCycleService_pb2.AddShockProfilesRequest(project=project)

        s = []
        profile = None
        for s in shock_profiles:
            profile = request.shockProfiles.add()
            profile.phaseName = s[0]
            profile.eventName = s[1]
            profile.profileName = s[2]
            profile.duration = s[3]
            profile.durationUnits = s[4]
            profile.sampleRate = s[5]
            profile.sampleRateUnits = s[6]
            profile.loadUnits = s[7]
            profile.freqUnits = s[8]

        # Add shock entries to the request
        for e in s[9]:
            entry = profile.shockProfileEntries.add()
            entry.shape = e[0]
            entry.load = e[1]
            entry.freq = e[2]
            entry.decay = e[3]

        response = self.stub.addShockProfiles(request)
        return_code = response.returnCode
        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockAddShockProfilesError(error_array=response.errors)
                else:
                    raise SherlockAddShockProfilesError(message=return_code.message)

            return return_code.value
        except SherlockAddShockProfilesError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def load_random_vibe_profile(
        self, project: str, phase_name: str, event_name: str, file_path: str
    ) -> int:
        """Load random vibe profile from .csv or .dat file.

        Available Since: 2023R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project
        phase_name: str
            Name of the lifecycle phase to add this event to.
        event_name: str
            Name of the random vibe event.
        file_path: str
            File path for thermal profile .dat or .csv file

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Example
        -------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card"
        )

        >>> sherlock.lifecycle.load_random_vibe_profile(
                project="Tutorial",
                phase_name="Phase 1",
                event_name="Random Event",
                file_path="TestProfile.dat"
        )
        """
        try:
            if project == "":
                raise SherlockLoadRandomVibeProfileError(message="Project name is invalid.")
            if phase_name == "":
                raise SherlockLoadRandomVibeProfileError(message="Phase name is invalid.")
            if event_name == "":
                raise SherlockLoadRandomVibeProfileError(message="Event name is invalid.")
            if file_path == "":
                raise SherlockLoadRandomVibeProfileError(message="File path is invalid.")
            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLifeCycleService_pb2.LoadRandomVibeProfileRequest(
                project=project,
                phaseName=phase_name,
                eventName=event_name,
                filePath=file_path,
            )
            response = self.stub.loadRandomVibeProfile(request)
            return_code = response.returnCode
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockLoadRandomVibeProfileError(error_array=response.errors)

                raise SherlockLoadRandomVibeProfileError(message=return_code.message)

            return return_code.value
        except SherlockLoadRandomVibeProfileError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def load_thermal_profile(
        self, project: str, phase_name: str, event_name: str, file_path: str
    ) -> int:
        """Load a thermal profile from a .dat or .csv file.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project
        phase_name: str
            Name of the lifecycle phase to add this event to.
        event_name: str
            Name of the random vibe event.
        file_path: str
            File path for thermal profile .dat or .csv file

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Example
        -------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card",
        )
         >>>loaded = sherlock.lifecycle.load_thermal_profile(
                project="Tutorial",
                phase_name="Phase 1",
                event_name="Thermal Event",
                file_path="Tutorial_Profile.dat"
        )
        """
        try:
            if project == "":
                raise SherlockLoadThermalProfileError(message="Project name is invalid.")
            if phase_name == "":
                raise SherlockLoadThermalProfileError(message="Phase name is invalid.")
            if event_name == "":
                raise SherlockLoadThermalProfileError(message="Event name is invalid.")
            if file_path == "":
                raise SherlockLoadThermalProfileError(message="File path is invalid.")
            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLifeCycleService_pb2.LoadThermalProfileRequest(
                project=project,
                phaseName=phase_name,
                eventName=event_name,
                filePath=file_path,
            )
            response = self.stub.loadThermalProfile(request)
            return_code = response.returnCode

            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockLoadThermalProfileError(error_array=response.errors)

                raise SherlockLoadThermalProfileError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockLoadThermalProfileError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version()
    def load_harmonic_profile(
        self, project: str, phase_name: str, event_name: str, file_path: str
    ) -> int:
        """Load a harmonic profile from a DAT or CSV file to a life cycle phase.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project
        phase_name: str
            Name of the life cycle phase to add the harmonic profile to.
        event_name: str
            Name of the harmonic event.
        file_path: str
            Path for DAT or CSV file with the harmonic profile.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Example
        -------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card"
        )

        >>> loaded = sherlock.lifecycle.load_harmonic_profile(
                project="Tutorial",
                phase_name="Phase 1",
                event_name="Harmonic Event",
                file_path="Test_Profile.dat"
        )
        """
        try:
            if project == "":
                raise SherlockLoadHarmonicProfileError(message="Project name is invalid.")
            if phase_name == "":
                raise SherlockLoadHarmonicProfileError(message="Phase name is invalid.")
            if event_name == "":
                raise SherlockLoadHarmonicProfileError(message="Event name is invalid.")
            if file_path == "":
                raise SherlockLoadHarmonicProfileError(message="File name is invalid.")
            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLifeCycleService_pb2.LoadHarmonicProfileRequest(
                project=project,
                phaseName=phase_name,
                eventName=event_name,
                filePath=file_path,
            )
            response = self.stub.loadHarmonicProfile(request)
            return_code = response.returnCode

            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockLoadHarmonicProfileError(error_array=response.errors)

                raise SherlockLoadHarmonicProfileError(message=return_code.message)

            return return_code.value
        except SherlockLoadHarmonicProfileError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def load_shock_profile_dataset(
        self, project: str, phase_name: str, event_name: str, file_path: str
    ) -> int:
        """Load shock profile dataset from a .csv or .dat file.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project
        phase_name: str
            Name of the lifecycle phase to add this event to.
        event_name: str
            Name of the random vibe event.
        file_path: str
            File path for thermal profile .dat or .csv file

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Example
        -------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card"
        )

        """
        try:
            if project == "":
                raise SherlockLoadShockProfileDatasetError(message="Project name is invalid.")
            if phase_name == "":
                raise SherlockLoadShockProfileDatasetError(message="Phase name is invalid.")
            if event_name == "":
                raise SherlockLoadShockProfileDatasetError(message="Event name is invalid.")
            if file_path == "":
                raise SherlockLoadShockProfileDatasetError(message="File path is invalid.")
            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLifeCycleService_pb2.LoadShockProfilePulsesRequest(
                project=project,
                phaseName=phase_name,
                eventName=event_name,
                filePath=file_path,
            )
            response = self.stub.loadShockProfileDataset(request)
            return_code = response.returnCode
            if return_code.value == -1:
                raise SherlockLoadShockProfileDatasetError(return_code.message)

            return return_code.value
        except SherlockLoadShockProfileDatasetError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def load_shock_profile_pulses(
        self, project: str, phase_name: str, event_name: str, file_path: str
    ) -> int:
        """Load shock profile pulses from a .csv .dat file.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project
        phase_name: str
            Name of the lifecycle phase to add this event to.
        event_name: str
            Name of the random vibe event.
        file_path: str
            Path for thermal profile .dat or .csv file

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Example
        -------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> sherlock = launch_sherlock()
        >>> sherlock.project.import_odb_archive(
            "ODB++ Tutorial.tgz",
            True,
            True,
            True,
            True,
            project="Test",
            cca_name="Card",
        )
        >>> sherlock.lifecycle.load_shock_profile_pulses(
                project="Tutorial",
                phase_name="Phase 1",
                event_name="Shock Event",
                file_path="Test_Profile.dat"
        )

        """
        try:
            if project == "":
                raise SherlockLoadShockProfilePulsesError(message="Project name is invalid.")
            if phase_name == "":
                raise SherlockLoadShockProfilePulsesError(message="Phase name is invalid.")
            if event_name == "":
                raise SherlockLoadShockProfilePulsesError(message="Event name is invalid.")
            if file_path == "":
                raise SherlockLoadShockProfilePulsesError(message="File path is invalid.")
            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLifeCycleService_pb2.LoadShockProfilePulsesRequest(
                project=project,
                phaseName=phase_name,
                eventName=event_name,
                filePath=file_path,
            )
            response = self.stub.loadShockProfilePulses(request)
            return_code = response.returnCode
            if return_code.value == -1:
                raise SherlockLoadShockProfilePulsesError(return_code.message)

            return return_code.value
        except SherlockLoadShockProfilePulsesError as e:
            LOG.error(str(e))
            raise e



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# © 2023 ANSYS, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""PySherlock client library."""

# Version
# ------------------------------------------------------------------------------

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata  # type: ignore

__version__ = importlib_metadata.version(__name__.replace(".", "-"))
"""PySherlock version."""

# Ease import statements
# ------------------------------------------------------------------------------

from ansys.sherlock.core.pysherlock_logging import Logger

LOG = Logger("sherlock")
"""PySherlock logger."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for launching Sherlock locally or connecting to a local instance with gRPC."""
import errno
import os
import shlex
import socket
import subprocess
import time
from typing import Optional

from ansys.sherlock.core import LOG
from ansys.sherlock.core.errors import SherlockCannotUsePortError, SherlockConnectionError
from ansys.sherlock.core.sherlock import Sherlock
from ansys.sherlock.core.utils.cyberchannel import create_channel
from ansys.sherlock.core.utils.version_check import _EARLIEST_SUPPORTED_VERSION

ANSYS_GRPC_CERTIFICATES = "ANSYS_GRPC_CERTIFICATES"
LOCALHOST = "127.0.0.1"
SHERLOCK_DEFAULT_PORT = 9090
SHERLOCK_UDS_SERVICE = "sherlock-grpc"
sherlock_cmd_args = []


def _is_port_available(host: str = LOCALHOST, port: int = SHERLOCK_DEFAULT_PORT):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        try:
            sock.bind((host, port))
            return True
        except socket.error as e:
            if e.errno == errno.EADDRINUSE:
                raise SherlockCannotUsePortError(port, "Port is already in use")

            raise SherlockCannotUsePortError(port, str(e))


def launch_sherlock(
    host: str = LOCALHOST,
    port: int = SHERLOCK_DEFAULT_PORT,
    single_project_path: str = "",
    sherlock_command_args: str = "",
    year: Optional[int] = None,
    release_number: Optional[int] = None,
    transport_mode: str = "mtls",
    certs_dir: str = None,
    uds_dir: str = None,
    uds_id: str = None,
) -> Sherlock:
    r"""Launch Sherlock and start gRPC on a given host and port.

    Parameters
    ----------
    host: str, optional
        IP address to start gRPC on. The default is ``"127.0.0.1"``, which
        is the IP address for the local host.
    port: int, optional
        Port number for the connection.
    single_project_path : str, optional
        Path to the Sherlock project if invoking Sherlock in the single-project mode.
    sherlock_command_args : str, optional
        Additional command arguments for launching Sherlock.
    year: int, optional
        4-digit year of the Sherlock release to launch. If not provided,
        the latest installed version of Sherlock will be launched.
    release_number: int, optional
        Release number of Sherlock to launch. If not provided,
        the latest installed version of Sherlock will be launched.
    transport_mode : str, optional
        Transport mode to use:
            - "insecure" : unencrypted connection (default)
            - "mtls" : mutual TLS authentication
            - "uds" : Unix Domain Socket
            - "wnua" : Windows Named User Authentication
    certs_dir: str, optional
        Directory containing the mTLS certificates. Default is "./certs".
    uds_dir : str, optional
        Directory for the UDS socket file.
    uds_id : str, optional
        Optional ID for the UDS socket file.

    Returns
    -------
    Sherlock
        The instance of sherlock.

    Examples
    --------
    >>> from ansys.sherlock.core import launcher
    >>> launcher.launch_sherlock()

    >>> from ansys.sherlock.core import launcher
    >>> launcher.launch_sherlock(port=9092, year=2024, release_number=1)

    >>> from ansys.sherlock.core import launcher
    >>> project = "C:\\Default Projects Directory\\ODB++ Tutorial"
    >>> launcher.launch_sherlock(port=9092, single_project_path=project)

    >>> from ansys.sherlock.core import launcher
    >>> launcher.launch_sherlock(
    >>>     port=9092,
    >>>     transport_mode="mtls",
    >>>     certs_dir="C:\\path\\to\\certs"
    >>> )

    >>> from ansys.sherlock.core import launcher
    >>> launcher.launch_sherlock(
    >>>     port=9092,
    >>>     transport_mode="uds",
    >>>     uds_dir="C:\\path\\to\\uds",
    >>>     uds_id="custom_id"
    >>> )

    """
    try:
        _is_port_available(host, port)
    except Exception as e:
        LOG.error(str(e))
        raise e

    _server_version = None
    try:
        sherlock_launch_cmd, _server_version = _get_sherlock_exe_path(
            year=year, release_number=release_number
        )
        args = [sherlock_launch_cmd]

        if transport_mode not in ["insecure", "mtls", "uds", "wnua"]:
            args.append(f"-grpcPort={port}")

        # Add gRPC options
        if transport_mode in ["insecure", "mtls", "wnua"]:
            args.append(f"-grpcHost={host}")
            args.append(f"-grpcPort={port}")
        args.append(f"--transport-mode={transport_mode}")
        if transport_mode == "mtls":
            if not certs_dir:
                certs_dir = os.getenv(ANSYS_GRPC_CERTIFICATES, "./certs")
            args.append(f"--certs-dir={certs_dir}")
        if transport_mode == "uds":
            if uds_dir:
                args.append(f"--uds-dir={uds_dir}")
            if uds_id:
                args.append(f"--uds-id={uds_id}")

        # Add other options
        if single_project_path:
            args.append("-singleProject")
            args.append(single_project_path)
        if sherlock_command_args:
            args.extend(shlex.split(sherlock_command_args))
        LOG.info(f"Command arguments: {args}")
        subprocess.Popen(args)

        sherlock = connect_grpc_channel(
            port=port,
            server_version=_server_version,
            transport_mode=transport_mode,
            certs_dir=certs_dir,
            uds_dir=uds_dir,
            uds_id=uds_id,
        )

        # Check that the gRPC connection is up (timeout after 3 minutes).
        count = 0
        while sherlock.common.check() is False and count < 90:
            time.sleep(2)
            count = count + 1

        if sherlock.common.check() is False:
            raise SherlockConnectionError(message="Error starting gRPC service")

        # Check that the Sherlock client has finished loading (timeout after 5 minutes).
        count = 0
        while sherlock.common.is_sherlock_client_loading() is False and count < 150:
            time.sleep(2)
            count = count + 1

        return sherlock
    except Exception as e:
        LOG.error("Error encountered while starting or executing Sherlock, error = %s" + str(e))


def connect_grpc_channel(
    port: int = SHERLOCK_DEFAULT_PORT,
    server_version: Optional[int] = None,
    transport_mode: str = "mtls",
    certs_dir: str = None,
    uds_dir: str = None,
    uds_id: str = None,
) -> Sherlock:
    """Create a gRPC connection to a specified port and return the ``Sherlock`` connection object.

    The ``Sherlock`` connection object is used to invoke the APIs from their respective services.
    This can be used to connect to the Sherlock instance that is already running with the specified
    port.

    Parameters
    ----------
    port: int, optional
        Port number for the connection. Default is ``SHERLOCK_DEFAULT_PORT``.
    server_version: int, optional
        Version of Sherlock. Default is the newest version that is installed.
    transport_mode: str, optional
        Transport mode for the gRPC connection. Default is ``"mtls"``.
    certs_dir: str, optional
        Directory containing the mTLS certificates. Default is ``None``.
    uds_dir: str, optional
        Directory for the UDS socket file. Default is ``None``.
    uds_id: str, optional
        Optional ID for the UDS socket file. Default is ``None``.

    Returns
    -------
    Sherlock
        The instance of sherlock.
    """
    try:
        channel = create_channel(
            transport_mode=transport_mode,
            host=LOCALHOST,
            port=port,
            uds_service=SHERLOCK_UDS_SERVICE,
            uds_dir=uds_dir,
            uds_id=uds_id,
            certs_dir=certs_dir,
            grpc_options=[
                ("grpc.max_send_message_length", 100 * 1024 * 1024),
                ("grpc.max_receive_message_length", 100 * 1024 * 1024),
            ],
        )
        LOG.info(f"gRPC channel created successfully using transport mode: {transport_mode}")
        return Sherlock(channel, server_version)
    except Exception as e:
        LOG.error(f"Failed to create gRPC channel with mode '{transport_mode}': {e}")
        raise


def _get_base_ansys(
    year: Optional[int] = None, release_number: Optional[int] = None
) -> tuple[str, int]:
    supported_installed_versions = {
        env_key: path
        for env_key, path in os.environ.items()
        if env_key.startswith("AWP_ROOT") and os.path.isdir(path)
    }

    if year is not None and release_number is not None:
        try:
            year = _extract_sherlock_version_year(year)

            sherlock_version = int(f"{year}{release_number}")
            version_key = f"AWP_ROOT{sherlock_version}"
            if version_key in supported_installed_versions:
                return supported_installed_versions[version_key], sherlock_version
            else:
                raise ValueError(f"Sherlock {year} {release_number} is not installed.")
        except ValueError as e:
            LOG.error(f"Error extracting Sherlock version year: {e}")
            raise e

    for key in sorted(supported_installed_versions, reverse=True):
        ansys_version = _get_ansys_version_from_awp_root(key)
        sherlock_version = int(ansys_version)
        if ansys_version >= _EARLIEST_SUPPORTED_VERSION:
            return supported_installed_versions[key], sherlock_version

    raise ValueError("Could not find any installed version of Sherlock.")


def _get_ansys_version_from_awp_root(awp_root: str):
    if awp_root.find("AWP_ROOT") >= 0:
        return int(awp_root.replace("AWP_ROOT", ""))

    return ""


def _get_sherlock_exe_path(
    year: Optional[int] = None, release_number: Optional[int] = None
) -> tuple[str, int]:
    ansys_base, sherlock_version = _get_base_ansys(year=year, release_number=release_number)
    if not ansys_base:
        return "", 0
    if os.name == "nt":
        sherlock_bin = os.path.join(ansys_base, "sherlock", "SherlockClient.exe")
    else:
        sherlock_bin = os.path.join(ansys_base, "sherlock", "runSherlock")
    return sherlock_bin, sherlock_version


def _extract_sherlock_version_year(year: int) -> int:
    if 1000 <= year <= 9999:
        return year % 100
    raise ValueError("Year must be a 4-digit integer.")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# © 2023 ANSYS, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing types for the Layer Service."""


from typing import Union

from ansys.api.sherlock.v0 import SherlockLayerService_pb2
from pydantic import BaseModel, ValidationInfo, field_validator, model_validator
from typing_extensions import Self

from ansys.sherlock.core.types.common_types import basic_str_validator


class PolygonalShape(BaseModel):
    """Contains the properties for a polygonal shape."""

    points: list[tuple[float, float]]
    """points (length two tuples of the form (x, y)) : list[tuple[float, float]]"""
    rotation: float
    """rotation (in degrees) : float"""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.PolygonalShape:
        grpc_polygonal_shape = SherlockLayerService_pb2.PolygonalShape()
        for x, y in self.points:
            p = grpc_polygonal_shape.points.add()
            p.x = x
            p.y = y
        grpc_polygonal_shape.rotation = self.rotation
        return grpc_polygonal_shape


class RectangularShape(BaseModel):
    """Contains the properties for a rectangular shape."""

    length: float
    """length : float"""
    width: float
    """width : float"""
    center_x: float
    """x coordinate of center : float"""
    center_y: float
    """y coordinate of center : float"""
    rotation: float
    """rotation (in degrees) : float"""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.RectangularShape:
        grpc_rectangular_shape = SherlockLayerService_pb2.RectangularShape()
        grpc_rectangular_shape.length = self.length
        grpc_rectangular_shape.width = self.width
        grpc_rectangular_shape.centerX = self.center_x
        grpc_rectangular_shape.centerY = self.center_y
        grpc_rectangular_shape.rotation = self.rotation
        return grpc_rectangular_shape


class SlotShape(BaseModel):
    """Contains the properties for a slot shape."""

    length: float
    """length : float"""
    width: float
    """width : float"""
    node_count: int
    """node count : int"""
    center_x: float
    """x coordinate of center : float"""
    center_y: float
    """y coordinate of center : float"""
    rotation: float
    """rotation (in degrees) : float"""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.SlotShape:
        grpc_slot_shape = SherlockLayerService_pb2.SlotShape()

        grpc_slot_shape.length = self.length
        grpc_slot_shape.width = self.width
        grpc_slot_shape.nodeCount = self.node_count
        grpc_slot_shape.centerX = self.center_x
        grpc_slot_shape.centerY = self.center_y
        grpc_slot_shape.rotation = self.rotation

        return grpc_slot_shape


class CircularShape(BaseModel):
    """Contains the properties for a circular shape."""

    diameter: float
    """diameter : float"""
    node_count: int
    """node count : int"""
    center_x: float
    """x coordinate of center : float"""
    center_y: float
    """y coordinate of center : float"""
    rotation: float
    """rotation (in degrees) : float"""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.CircularShape:
        grpc_circular_shape = SherlockLayerService_pb2.CircularShape()
        grpc_circular_shape.diameter = self.diameter
        grpc_circular_shape.nodeCount = self.node_count
        grpc_circular_shape.centerX = self.center_x
        grpc_circular_shape.centerY = self.center_y
        grpc_circular_shape.rotation = self.rotation
        return grpc_circular_shape


class PCBShape(BaseModel):
    """Contains the properties for a PCB shape."""

    @staticmethod
    def _convert_to_grpc() -> SherlockLayerService_pb2.PCBShape:
        return SherlockLayerService_pb2.PCBShape()


class PottingRegion(BaseModel):
    """Contains the properties of a Potting Region add or update request."""

    cca_name: str
    """The name of the CCA."""
    potting_id: str
    """The potting ID."""
    potting_side: str
    """The potting side, options are "TOP", "BOT", or "BOTTOM"."""
    potting_material: str
    """The potting material."""
    potting_units: str
    """The units to use for the potting region."""
    potting_thickness: float
    """The potting thickness."""
    potting_standoff: float
    """The potting standoff."""
    shape: Union[CircularShape, PCBShape, PolygonalShape, RectangularShape, SlotShape]
    """The shape of the potting region."""

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.PottingRegion:
        grpc_potting_region_data = SherlockLayerService_pb2.PottingRegion()

        grpc_potting_region_data.ccaName = self.cca_name
        grpc_potting_region_data.pottingID = self.potting_id
        grpc_potting_region_data.pottingSide = self.potting_side
        grpc_potting_region_data.pottingMaterial = self.potting_material
        grpc_potting_region_data.pottingUnits = self.potting_units
        grpc_potting_region_data.pottingThickness = self.potting_thickness
        grpc_potting_region_data.pottingStandoff = self.potting_standoff

        if isinstance(self.shape, CircularShape):
            grpc_potting_region_data.circularShape.CopyFrom(self.shape._convert_to_grpc())
        elif isinstance(self.shape, PCBShape):
            grpc_potting_region_data.pCBShape.CopyFrom(self.shape._convert_to_grpc())
        elif isinstance(self.shape, PolygonalShape):
            grpc_potting_region_data.polygonalShape.CopyFrom(self.shape._convert_to_grpc())
        elif isinstance(self.shape, RectangularShape):
            grpc_potting_region_data.rectangularShape.CopyFrom(self.shape._convert_to_grpc())
        elif isinstance(self.shape, SlotShape):
            grpc_potting_region_data.slotShape.CopyFrom(self.shape._convert_to_grpc())
        else:
            raise ValueError("Unsupported shape given '" + type(self.shape).__name__ + "'")

        return grpc_potting_region_data

    @field_validator("cca_name", "potting_id", "potting_side", "potting_material", "potting_units")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class PottingRegionUpdateData(BaseModel):
    """Contains the properties of a potting region update request."""

    potting_region_id_to_update: str
    """Id of the potting region to update."""
    potting_region: PottingRegion
    """Potting region data used to update the potting region."""

    def _convert_to_grpc(
        self,
    ) -> SherlockLayerService_pb2.UpdatePottingRegionRequest.PottingRegionUpdateData:

        grpc_potting_update_data = (
            SherlockLayerService_pb2.UpdatePottingRegionRequest.PottingRegionUpdateData()
        )

        grpc_potting_update_data.pottingRegionIDToUpdate = self.potting_region_id_to_update
        grpc_potting_update_data.pottingRegion.CopyFrom(self.potting_region._convert_to_grpc())
        return grpc_potting_update_data

    @field_validator("potting_region_id_to_update")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class UpdatePottingRegionRequest(BaseModel):
    """Contains the properties of a potting region update per project."""

    project: str
    """Name of the Sherlock project."""
    update_potting_regions: list[PottingRegionUpdateData]
    """List of potting region data to update."""

    @field_validator("project")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(
        self,
    ) -> SherlockLayerService_pb2.UpdatePottingRegionRequest:
        request = SherlockLayerService_pb2.UpdatePottingRegionRequest()
        request.project = self.project
        for update_potting_region in self.update_potting_regions:
            request.updatePottingRegions.append(update_potting_region._convert_to_grpc())
        return request


class PottingRegionCopyData(BaseModel):
    """Data identifying which potting regions to copy."""

    cca_name: str
    """Name of the cca."""
    potting_id: str
    """Id to assign to the new potting region."""
    copy_potting_id: str
    """Id of the potting region to copy."""
    center_x: float
    """X coordinate for the center of the new potting region."""
    center_y: float
    """Y coordinate for the center of the new potting region."""

    @field_validator("cca_name", "potting_id", "copy_potting_id")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @model_validator(mode="after")
    def validate_ids(self) -> Self:
        """Validate that the potting IDs are not the same between the original and copy."""
        if self.potting_id == self.copy_potting_id:
            raise ValueError("potting_id must be different than copy_potting_id")
        return self

    def _convert_to_grpc(
        self,
    ) -> SherlockLayerService_pb2.CopyPottingRegionRequest.PottingRegionCopyData:

        grpc_potting_copy_data = (
            SherlockLayerService_pb2.CopyPottingRegionRequest.PottingRegionCopyData()
        )

        grpc_potting_copy_data.ccaName = self.cca_name
        grpc_potting_copy_data.pottingID = self.potting_id
        grpc_potting_copy_data.copyPottingID = self.copy_potting_id
        grpc_potting_copy_data.centerX = self.center_x
        grpc_potting_copy_data.centerY = self.center_y

        return grpc_potting_copy_data


class CopyPottingRegionRequest(BaseModel):
    """Request to delete 1 or more potting regions."""

    project: str
    """Name of the project containing the potting region(s) to be copied."""
    potting_region_copy_data: list[PottingRegionCopyData]
    """Data identifying which potting regions to copy and what potting regions to copy from."""

    @field_validator("project")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(
        self,
    ) -> SherlockLayerService_pb2.CopyPottingRegionRequest:

        grpc_potting_copy_request = SherlockLayerService_pb2.CopyPottingRegionRequest()
        grpc_potting_copy_request.project = self.project
        for copy_data in self.potting_region_copy_data:
            grpc_potting_copy_request.pottingRegionCopyData.append(copy_data._convert_to_grpc())

        return grpc_potting_copy_request


class PottingRegionDeleteData(BaseModel):
    """Data specifying potting regions to delete."""

    cca_name: str
    """Name of the CCA containing the potting region(s) to delete."""
    potting_id: str
    """Id of the potting region(s) to delete."""

    @field_validator("cca_name", "potting_id")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(
        self,
    ) -> SherlockLayerService_pb2.DeletePottingRegionRequest.PottingRegionDeleteData:
        delete_data = SherlockLayerService_pb2.DeletePottingRegionRequest.PottingRegionDeleteData()

        delete_data.ccaName = self.cca_name
        delete_data.pottingID = self.potting_id
        return delete_data


class DeletePottingRegionRequest(BaseModel):
    """Request to delete 1 or more potting regions."""

    project: str
    """Name of the project containing the potting regions to delete."""
    potting_region_delete_data: list[PottingRegionDeleteData]
    """Data identifying which potting regions should be deleted."""

    @field_validator("project")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> SherlockLayerService_pb2.DeletePottingRegionRequest:
        request = SherlockLayerService_pb2.DeletePottingRegionRequest()
        request.project = self.project
        for delete_data in self.potting_region_delete_data:
            request.pottingRegionDeleteData.append(delete_data._convert_to_grpc())
        return request



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing types for the Parts Service."""

import warnings

try:
    import SherlockCommonService_pb2
    import SherlockPartsService_pb2
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockCommonService_pb2, SherlockPartsService_pb2


def deprecation(cls: object):
    """Raise a DeprecationWarning when a deprecated class is used."""
    # if str(cls).find("PartsListSearchMatchingMode") != -1:
    message = f"{cls} is deprecated. Use a string with the value of the constant name \
    as defined in the proto file."
    warnings.warn(message, DeprecationWarning, stacklevel=2)
    return cls


@deprecation
class PartsListSearchMatchingMode:
    """DEPRECATED. Constants for Matching Mode in Update Parts List & Update Parts from AVL."""

    matching_mode = SherlockCommonService_pb2.MatchingMode
    BOTH = matching_mode.Both
    """Both"""
    PART = matching_mode.Part
    """Part"""


class PartsListSearchDuplicationMode:
    """Constants for Duplication Mode in Update Parts List and Update Parts from AVL request."""

    duplication_mode = SherlockPartsService_pb2.DuplicationMode
    FIRST = duplication_mode.First
    """First"""
    ERROR = duplication_mode.Error
    """Error"""
    IGNORE = duplication_mode.Ignore
    """Ignore"""


class AVLPartNum:
    """Constants for AVLPartNum in the Update Parts List from AVL request."""

    avl_part_num = SherlockPartsService_pb2.AVLPartNum
    ASSIGN_INTERNAL_PART_NUM = avl_part_num.AssignInternalPartNum
    """AssignInternalPartNum"""
    ASSIGN_VENDOR_AND_PART_NUM = avl_part_num.AssignVendorAndPartNum
    """AssignVendorAndPartNum"""
    DO_NOT_CHANGE_VENDOR_OR_PART_NUM = avl_part_num.DoNotChangeVendorOrPartNum
    """DoNotChangeVendorOrPartNum"""


class AVLDescription:
    """Constants for AVLDescription in the Update Parts List from AVL request."""

    avl_description = SherlockPartsService_pb2.AVLDescription
    ASSIGN_APPROVED_DESCRIPTION = avl_description.AssignApprovedDescription
    """AssignApprovedDescription"""
    DO_NOT_CHANGE_DESCRIPTION = avl_description.DoNotChangeDescription
    """DoNotChangeDescription"""


class PartLocation:
    """Part Location property values."""

    def __init__(self, location):
        """Initialize members from the location."""
        self.x = location.x
        """x coordinate"""
        self.y = location.y
        """y coordinate"""
        self.rotation = location.rotation
        """rotation (in degrees)"""
        self.location_units = location.locationUnits
        """units for location coordinates"""
        self.board_side = location.boardSide
        """board side - ``"TOP"`` or ``"BOTTOM"`` """
        self.mirrored = location.mirrored
        """mirrored - ``True`` or ``False`` """
        self.ref_des = location.refDes
        """reference designator"""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing all model generation capabilities."""
import os.path

import grpc

from ansys.sherlock.core.types.analysis_types import ElementOrder

try:
    import SherlockModelService_pb2
    from SherlockModelService_pb2 import MeshType, TraceOutputType
    import SherlockModelService_pb2_grpc
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockModelService_pb2
    from ansys.api.sherlock.v0.SherlockModelService_pb2 import MeshType, TraceOutputType
    from ansys.api.sherlock.v0 import SherlockModelService_pb2_grpc

from ansys.sherlock.core import LOG
from ansys.sherlock.core.errors import (
    SherlockExportAEDBError,
    SherlockExportFEAModelError,
    SherlockModelServiceError,
    SherlockNoGrpcConnectionException,
)
from ansys.sherlock.core.grpc_stub import GrpcStub
from ansys.sherlock.core.types.common_types import Measurement
from ansys.sherlock.core.utils.version_check import require_version


class Model(GrpcStub):
    """Contains all model generation capabilities."""

    def __init__(self, channel: grpc.Channel, server_version: int):
        """Initialize a gRPC stub for the Sherlock Model service."""
        super().__init__(channel, server_version)
        self.stub = SherlockModelService_pb2_grpc.SherlockModelServiceStub(channel)

    @require_version()
    def export_trace_reinforcement_model(
        self,
        project_name: str,
        cca_name: str,
        export_file: str,
        overwrite: bool = True,
        display_model: bool = False,
        generate_models_for_all_layers: bool = False,
        coordinate_units: str = "mm",
        trace_param_diameter_threshold_val: float = 2,
        trace_param_diameter_threshold_unit: str = "mm",
        trace_param_min_hole_diameter_val: float = 0.25,
        trace_param_min_hole_diameter_unit: str = "mm",
        trace_drill_hole_modeling: str = "DISABLED",
        trace_drill_hole_min_diameter_val: float = 2,
        trace_drill_hole_min_diameter_unit: str = "mm",
        trace_drill_hole_max_edge_val: float = 1,
        trace_drill_hole_max_edge_unit: str = "mm",
    ) -> int:
        r"""Export a trace reinforcement model.

        Available Since: 2023R1

        Parameters
        ----------
        project_name: str
            Name of the Sherlock project to generate the trace reinforcement model for.
        cca_name: str
            Name of the CCA to generate the trace reinforcement model from.
        export_file: str
            Path for saving exported files to. The file extension must be ``.wbjn``.
        overwrite: bool, optional
            Whether to overwrite an existing file having the same file name.
            The default is ``True``.
        display_model: bool, optional
            Whether to launch and display the exported model in Ansys Workbench
            Mechanical once the export finishes. The default is ``False``.
        generate_models_for_all_layers:  bool, optional
            Whether to generate and export trace models for not only the generated trace
            reinforcement layers but also all other layers. The default is ``False``, in
            which case only trace reinforcement layers are generated and exported.
        coordinate_units: str, optional
            Units of the model coordinates to use when exporting a model.
            The default is ``"mm"``.
        trace_param_diameter_threshold_val: float, optional
            Threshold value that determines whether a hole is modeled with shell
            reinforcement elements or beam elements. The default is ``2``, with the
            default units being ``"mm"`` as specified by the next parameter. Holes with
            diameters equal to or greater than this threshold value are modeled with shell
            reinforcement elements. Holes with diameters less than this threshold value
            are modeled with beam elements. Holes buried inside the board are always modeled
            with beam elements.
        trace_param_diameter_threshold_unit: str, optional
            Units associated with the threshold value for the trace parameter diameter.
            The default is ``"mm"``.
        trace_param_min_hole_diameter_val: float, optional
            Minimum trace parameter diameter for determining whether a via is exported.
            The default is ``0.25``, with the default units being ``"mm"`` as specified
            by the next parameter. Vias with diameters smaller than this diameter
            are not exported. Setting the value to ``0`` exports all vias.
        trace_param_min_hole_diameter_unit: str, optional
            Units associated with the value for the minimum trace parameter diameter.
            The default is ``"mm"``.
        trace_drill_hole_modeling: str, optional
            Whether to enable or disable the modeling of trace drill holes. Options are
            ``"ENABLED"`` and ``"DISABLED"``. The default is ``"DISABLED"``, in which
            case the ``trace_drill_hole_min_diameter`` and ``trace_drill_hole_max_edge``
            parameters are not used.
        trace_drill_hole_min_diameter_val: float, optional
            Minimimun diameter value for determining whether a trace drill hole is
            exported. The default is ``2``, with the default units being ``"mm"``
            as specified by the next parameter. Trace drill holes with diameters smaller
            than this diameter are not exported. Setting the value to ``0`` exports all
            trace drill holes.
        trace_drill_hole_min_diameter_unit: str, optional
            Units associated with the value for the minimum trace drill hole diameter.
            The default is ``"mm"``.
        trace_drill_hole_max_edge_val: float, optional
            Maximum segment size for representing round drill holes by a polygon.
            The default is ``1``, with the default units being ``"mm"`` as specified
            by the next parameter.
        trace_drill_hole_max_edge_unit: str, optional
            Units associated with the maximum segment for representing round drill holes
            by a polygon. The default is ``"mm"``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> from ansys.sherlock.core import model
        >>> sherlock = launcher.launch_sherlock()
        >>> sherlock.model.export_trace_reinforcement_model(
            'Tutorial Project', 'Main Board', 'c:\Temp\export.wbjn',
            True, False, False)

        >>> from ansys.sherlock.core import launcher
        >>> from ansys.sherlock.core import model
        >>> sherlock = launcher.launch_sherlock()
        >>> sherlock.model.export_trace_reinforcement_model(
            'Tutorial Project', 'Main Board', 'c:\Temp\export.wbjn',
            True, False, False, "mm", 1.5, "mm", 0, "mm", "ENABLED", 1.5, "mm", 1, "mm")
        """
        try:
            if not project_name:
                raise SherlockModelServiceError("Project name is invalid.")
            if not cca_name:
                raise SherlockModelServiceError("CCA name is invalid.")
            if export_file == "":
                raise SherlockModelServiceError(message="Export file path is invalid.")
            else:
                if not os.path.exists(os.path.dirname(export_file)):
                    raise SherlockModelServiceError(
                        message=f'Export file directory "{export_file}" does not exist.'
                    )
        except Exception as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        export_request = SherlockModelService_pb2.ExportTraceReinforcementModelRequest()
        export_request.project = project_name
        export_request.ccaName = cca_name
        export_request.exportFile = export_file
        export_request.overwrite = overwrite
        export_request.displayModel = display_model
        export_request.clearFEADatabase = (
            False  # This only applies to *.apdl files and is not applicable here.
        )
        export_request.coordinateUnits = coordinate_units
        export_request.generateModelsForAllLayers = generate_models_for_all_layers
        export_request.traceParam.diameterThreshold.value = trace_param_diameter_threshold_val
        export_request.traceParam.diameterThreshold.unit = trace_param_diameter_threshold_unit
        export_request.traceParam.minHoleDiameterForShellOrBeam.value = (
            trace_param_min_hole_diameter_val
        )
        export_request.traceParam.minHoleDiameterForShellOrBeam.unit = (
            trace_param_min_hole_diameter_unit
        )

        # export_request.traceDrillHoleParam is Deprecated
        export_request.traceDrillHoleParam.drillHoleModeling = trace_drill_hole_modeling
        export_request.traceDrillHoleParam.minHoleDiameter.value = trace_drill_hole_min_diameter_val
        export_request.traceDrillHoleParam.minHoleDiameter.unit = trace_drill_hole_min_diameter_unit
        export_request.traceDrillHoleParam.maxEdgeLength.value = trace_drill_hole_max_edge_val
        export_request.traceDrillHoleParam.maxEdgeLength.unit = trace_drill_hole_max_edge_unit
        # export_request.traceDrillHoleParam is Deprecated

        # New message to replace traceDrillHoleParam below.
        # Convert ENABLE/ENABLED or DISABLE/DISABLED to boolean
        export_request.drillHoleModeling.drillHoleModelingEnabled = (
            trace_drill_hole_modeling.upper().startswith("ENABLE")
        )
        export_request.drillHoleModeling.minHoleDiameter.value = trace_drill_hole_min_diameter_val
        export_request.drillHoleModeling.minHoleDiameter.units = trace_drill_hole_min_diameter_unit
        export_request.drillHoleModeling.maxEdgeLength.value = trace_drill_hole_max_edge_val
        export_request.drillHoleModeling.maxEdgeLength.units = trace_drill_hole_max_edge_unit

        try:
            return_code = self.stub.exportTraceReinforcementModel(export_request)
            if return_code.value != 0:
                raise SherlockModelServiceError(return_code.message)

            return return_code.value
        except Exception as e:
            LOG.error(str(e))
            raise

    @require_version()
    def generate_trace_model(
        self,
        project_name,
        cca_name: str = "",
        copper_layer_name: str = "",
        max_arc_segment: float = 0.0,
        max_arc_segment_units: str = "mm",
        min_trace_area: float = 0.0,
        min_trace_area_units: str = "mm2",
        min_hole_area: float = 0.0,
        min_hole_area_units: str = "mm2",
        use_snapshot_for_non_image_layer: bool = False,
    ) -> int:
        r"""Generate one or more trace models for a project.

        Available Since: 2023R2

        Parameters
        ----------
        project_name: str
            Name of the Sherlock project to generate one or more trace models for.
        cca_name: str, optional
            Name of the CCA to generate one or more trace models from. The default is
            ``""``, in which case trace models are generated for CCAs and
            all layers.
        copper_layer_name: str, optional
            Name of the copper layer to generate one or more trace models from. The default
            is ``""``, in which case trace models are generated either for the given CCA
            or for all layers.
        max_arc_segment: float, optional
            Maximum length of the segment to generate when Sherlock
            converts EDA arc drawing commands to line segments. The default is
            ``0.0``. Smaller values for the maximum arc segment result in smoother
            arc representations on the FEA model. However, the cost of generating a
            larger number of shorter segments is higher. Such short segments cause
            the FEA tool to generate a larger number of smaller elements to represent
            the curved solid.
        max_arc_segment_units: str, optional
            Units for the maximum arc segment. The default is ``"mm"``.
        min_trace_area: float, optional
            Minimum area of any trace polygon to include in the trace model.
            The default is ``0.0``, which turns off any area filtering.
        min_trace_area_units: str, optional
            Units for the minimum trace area. The default is ``"mm2"``.
        min_hole_area: float, optional
            Minimum area of any trace hole to include in the trace model.
            The default is ``0.0``, which turns off any hole filtering.
        min_hole_area_units: str, optional
            Units for the minimum hole area. The default is ``"mm2"``.
        use_snapshot_for_non_image_layer: bool, optional
            Whether to use an image to generate the trace model for layers that are not
            image layers. The default is ``False``. If ``True`` and a snapshot image for
            the layer exists, the snapshot image is used. Otherwise, an image is created
            in the same way as a snapshot image is created.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> from ansys.sherlock.core import model
        >>> sherlock = launcher.launch_sherlock()
        >>> sherlock.model.generate_trace_model(
            'Tutorial Project', 'Main Board', 0.05, 'mm'
            0.0, 'mm2', 0.0, 'mm2')

        """
        try:
            if not project_name:
                raise SherlockModelServiceError("Project name is invalid.")

        except Exception as e:
            LOG.error(str(e))
            raise

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        gen_request = SherlockModelService_pb2.GenerateTraceModelRequest()
        gen_request.project = project_name
        gen_request.ccaName = cca_name
        gen_request.copperLayerName = copper_layer_name
        gen_request.maxArcSegment = max_arc_segment
        gen_request.maxArcSegmentUnits = max_arc_segment_units
        gen_request.minTraceArea = min_trace_area
        gen_request.minTraceAreaUnits = min_trace_area_units
        gen_request.minHoleArea = min_hole_area
        gen_request.minHoleAreaUnits = min_hole_area_units
        gen_request.useSnapshotForNonImageLayer = use_snapshot_for_non_image_layer

        try:
            return_code = self.stub.generateTraceModel(gen_request)
            if return_code.value != 0:
                raise SherlockModelServiceError(return_code.message)

            return return_code.value
        except Exception as e:
            LOG.error(str(e))
            raise

    @require_version(242)
    def export_aedb(
        self,
        project_name: str,
        cca_name: str,
        export_file: str,
        overwrite: bool = True,
        display_model: bool = False,
    ) -> int:
        r"""Export an Electronics Desktop model.

        Available Since: 2024R2

        Parameters
        ----------
        project_name: str
            Name of the Sherlock project to generate the EDB model for.
        cca_name: str
            Name of the CCA to generate the EDB model from.
        export_file: str
            Directory for saving exported model to.
        overwrite: bool, optional
            Whether to overwrite an existing file having the same file name.
            The default is ``True``.
        display_model: bool, optional
            Whether to launch and display the exported model in Ansys Electronics
            Desktop once the export finishes. The default is ``False``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> from ansys.sherlock.core import model
        >>> sherlock = launcher.launch_sherlock()
        >>> sherlock.model.export_aedb(
            'Tutorial Project', 'Main Board', 'c:\Temp\export.aedb',
            True, False)
        """
        try:
            if not project_name:
                raise SherlockExportAEDBError("Project name is invalid.")
            if not cca_name:
                raise SherlockExportAEDBError("CCA name is invalid.")
            if export_file == "":
                raise SherlockExportAEDBError(message="Export filepath is required.")
        except Exception as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        export_request = SherlockModelService_pb2.ExportAEDBRequest()
        export_request.project = project_name
        export_request.ccaName = cca_name
        export_request.exportFile = export_file
        export_request.overwrite = overwrite
        export_request.displayModel = display_model

        try:
            return_code = self.stub.exportAEDB(export_request)
            if return_code.value != 0:
                raise SherlockExportAEDBError(return_code.message)

            return return_code.value
        except Exception as e:
            LOG.error(str(e))
            raise

    @require_version(242)
    def exportTraceModel(self, layer_params: list[bool | int | float | str]) -> int:
        r"""Export a trace model to a specified output file.

        Available Since: 2024R2

        Parameters
        ----------
        layer_params : list[bool | int | float | str]
            list of parameters for export a trace model of a single copper layer.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.types.analysis_types import ElementOrder
        >>> from ansys.sherlock.core import launcher
        >>> from ansys.api.sherlock.v0 import SherlockModelService_pb2
        >>> sherlock = launcher.launch_sherlock()
        >>> list_of_params_for_layers = []
        >>> list_of_params_for_layers.append(
                sherlock.model.createExportTraceCopperLayerParams(
                    "Tutorial Project",
                    "Main Board",
                    ".\\outputfile_path.stp",
                    "copper-01.odb",
                    False,
                    False,
                    False,
                    False,
                    "mm",
                    SherlockModelService_pb2.MeshType.NONE,
                    False,
                    SherlockModelService_pb2.TraceOutputType.ALL_REGIONS,
                    ElementOrder.LINEAR,
                    1.0,
                    "mm".
                    1,
                    False,
                    1.0,
                    "mm",
                    1.0
                )
            )
        >>> sherlock.model.exportTraceModel(list_of_params_for_layers)
        """
        try:
            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockModelService_pb2.ExportTraceModelRequest()
            request.traceModelExportParams.extend(layer_params)

            return_code = self.stub.exportTraceModel(request)
            if return_code.value != 0:
                # Return error from the server
                raise SherlockModelServiceError(return_code.message)

            return return_code.value
        except Exception as e:
            LOG.error(str(e))
            raise

    @require_version(242)
    def createExportTraceCopperLayerParams(
        self,
        project_name,
        cca_name,
        output_file_path,
        copper_layer,
        overwrite: bool = False,
        display_after: bool = False,
        clear_FEA_database: bool = False,
        use_FEA_model_ID: bool = False,
        coord_units: str = "mm",
        mesh_type: int = MeshType.NONE,
        is_modeling_region_enabled: bool = False,
        trace_output_type: int = TraceOutputType.ALL_REGIONS,
        element_order: ElementOrder = ElementOrder.LINEAR,
        max_mesh_size: float = 1.0,
        max_mesh_size_units: str = "mm",
        max_holes_per_trace: int = 2,
        is_drill_hole_modeling_enabled: bool = False,
        drill_hole_min_diameter: float = 1.0,
        drill_hole_min_diameter_units: str = "mm",
        drill_hole_max_edge_length: float = 1.0,
        drill_hole_max_edge_length_units: str = "mm",
    ) -> SherlockModelService_pb2.TraceModelExportParams:
        r"""Create a set of parameters to be used to export a single copper layer.

        Creates TraceModelExportParams object that can be added to an export trace model request.
        Should be used in conjunction with exportTraceModel method to export multiple trace layers
        all at once. See example below.

        Parameters
        ----------
        project_name: str
            Name of the Sherlock project containing trace layer to export.
        cca_name: str
            Name of the CCA containing the trace layer to export.
        output_file_path: str
            File path including the file name and extension where the trace layer will be exported.
            Valid file extensions: .py, .bdf, .apdl, .cdb, .wbjn, .stp, .step, .tcl, .stl
            Note: relative paths will be relative to sherlock install directory,
            not the python script.
        copper_layer: str
            Name of the copper layer in the given CCA to export.
        overwrite: bool = False
            Determines if sherlock should overwrite the output file if it exists.
        display_after: bool = False
            Determines if the output file should automatically display after export.
        clear_FEA_database: bool = False
            Determines if sherlock should clear the database after export.
            Applicable file extensions: .apdl, and .cdb.
        use_FEA_model_ID: bool = False
            Determines if the FEA model id is used or not.
        coord_units: str = "mm"
            Units of the coordinate system. Applicable to .py .wbjn, .stp, .step.
        mesh_type: MeshType = MeshType.NONE
            Options of difference trace meshing strategies
        is_modeling_region_enabled: bool = False
            Determines if pre-defined modeling regions will be applied to the exported trace model.
        trace_output_type: TraceOutputType = TraceOutputType.ALL_REGIONS
            Options to select which trace regions to include in the 3D model.
        element_order: ElementOrder = ElementOrder.LINEAR
            Type of FEA element to be used when modeling each component.
        max_mesh_size: float = 1.0
            Indicates the desired element sizes.
        max_mesh_size_units: str = "mm"
            Indicates the units to be used with max_mesh_size.
        max_holes_per_trace: int = 2
            Maximum number of holes allowed in a trace before partitioning it into multiple traces.
        is_drill_hole_modeling_enabled: bool = False
            Determines if drill holes will be modeled or not.
        drill_hole_min_diameter: float = 1.0
            All drill holes with a diameter < this value will not be modeled.
        drill_hole_min_diameter_units : str = "mm"
            Units associated with drill_hole_min_diameter.
        drill_hole_max_edge_length: float = 1.0
            Specifies the length of the line segments used to represent round drill holes.
        drill_hole_max_edge_length_units: str = "mm"
            Units associated with drill_hole_max_edge_length.

        Returns
        -------
        TraceModelExportParams
            Object that holds the data for a single export trace request.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> from ansys.sherlock.core.types.analysis_types import ElementOrder
        >>> from ansys.api.sherlock.v0 import SherlockModelService_pb2
        >>> sherlock = launcher.launch_sherlock()
        >>> copper_1_layer = sherlock.model.createExportTraceCopperLayerParams(
                "Tutorial Project",
                "Main Board",
                ".\\outputfile_path.stp",
                "copper-01.odb",
                False,
                False,
                False,
                False,
                "mm",
                SherlockModelService_pb2.MeshType.NONE,
                False,
                SherlockModelService_pb2.TraceOutputType.ALL_REGIONS,
                ElementOrder.LINEAR,
                1.0,
                "mm",
                2,
                False,
                1.0,
                "mm",
                1.0
            )
        >>> copper_2_layer = sherlock.model.createExportTraceCopperLayerParams(
                "Tutorial Project",
                "Main Board",
                ".\\outputfile_path2.stp",
                "copper-02.odb",
                False,
                False,
                False,
                False,
                "mm",
                SherlockModelService_pb2.MeshType.NONE,
                False,
                SherlockModelService_pb2.TraceOutputType.ALL_REGIONS,
                ElementOrder.LINEAR,
                1.0,
                "mm",
                2,
                False,
                1.0,
                "mm",
                1.0
            )
        >>> sherlock.model.exportTraceModel([copper_1_layer, copper_2_layer])
        """
        try:
            if not project_name:
                raise SherlockModelServiceError("Project name is invalid.")
            if not cca_name:
                raise SherlockModelServiceError("CCA name is invalid.")
            if not output_file_path:
                raise SherlockModelServiceError("Output File path is required")
            if not copper_layer:
                raise SherlockModelServiceError("Copper layer name is required.")
        except Exception as e:
            LOG.error(str(e))
            raise

        ret = SherlockModelService_pb2.TraceModelExportParams()

        ret.project = project_name
        ret.ccaName = cca_name
        ret.filePath = output_file_path
        ret.copperLayerName = copper_layer
        ret.overwriteExistingFile = overwrite
        ret.displayModelAfterExport = display_after
        ret.clearFEADatabase = clear_FEA_database
        ret.useFEAModelID = use_FEA_model_ID
        ret.coordUnits = coord_units

        # Mesh Type Params
        pmp = ret.pcbMeshPropParam
        pmp.meshType = mesh_type
        pmp.isModelingRegionEnabled = is_modeling_region_enabled

        # Trace Params
        tpp = ret.tracePropParam
        tpp.traceOutputs = trace_output_type
        tpp.elementOrder = element_order
        tpp.maxMeshSize.value = max_mesh_size
        tpp.maxMeshSize.units = max_mesh_size_units
        tpp.maxHolesPerTrace = max_holes_per_trace

        # Drill Hole Params
        dhm = ret.drillHoleModeling
        dhm.drillHoleModelingEnabled = is_drill_hole_modeling_enabled
        dhm.minHoleDiameter.value = drill_hole_min_diameter
        dhm.minHoleDiameter.units = drill_hole_min_diameter_units
        dhm.maxEdgeLength.value = drill_hole_max_edge_length
        dhm.maxEdgeLength.units = drill_hole_max_edge_length_units

        return ret

    def export_FEA_model(
        self,
        project: str,
        cca_name: str,
        export_file: str,
        analysis: str,
        drill_hole_parameters: list[dict[str, str | Measurement]],
        detect_lead_modeling: str,
        lead_model_parameters: list[dict[str, int | str | Measurement]],
        display_model: bool,
        clear_FEA_database: bool,
        use_FEA_model_id: bool,
        coordinate_units: str,
    ) -> int:
        """
        Export a FEA model.

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        export_file: str
            Full path for saving exported files to. The file extension must be ``.wbjn``.
        analysis: str
            The type of analysis that is being exported. Valid values are ``NaturalFreq``,
            ``HarmonicVibe``, ``ICTAnalysis``, ``MechanicalShock`` or ``RandomVibe``.
        drill_hole_parameters: list[dict[str, str | Measurement]]
            List of the drill hole parameters consisting of these properties:

                - drill_hole_modeling: str
                    The status of the drill hole modeling feature. If enabled, automatically enable
                    drill hole modeling. Valid values are ``ENABLED/enabled`` or
                    ``DISABLED/disabled``.
                - min_hole_diameter: MinHoleDiameter
                    The properties of the minimum hole diameter.
                - max_edge_length: MaxEdgeLength
                    The properties of the maximum edge length.
        detect_lead_modeling: str
            The status of the detect lead modeling feature. If enabled, automatically enable lead
            modeling if any part has lead geometry defined. Valid values are ``ENABLED`` or
            ``DISABLED``.
        lead_model_parameters: list[dict[str, int | str | Measurement]]
            List of the lead model parameters consisting of these properties:

                - lead_modeling: str
                    The status of the lead modeling feature. If enabled, automatically enable lead
                    modeling. Valid values are ``ENABLED`` or ``DISABLED``.
                - lead_element_order: str
                     The type of the element order. Valid values are ``First Order (Linear)``,
                     ``Second Order (Quadratic)``, or ``Solid Shell``.
                - max_mesh_size: MaxMeshSize
                    The properties of the maximum mesh size.
                - vertical_mesh_size: VerticalMeshSize
                    The properties of the vertical mesh size.
                - thicknessCount: int, optional
                    The number of elements through the lead thickness that will be created per lead.
                     The default value is 3 and the maximum is 5. Only used when the advanced lead
                     mesh setting is enabled.
                - aspectRatio: int, optional
                    The aspect ratio is multiplied by the lead thickness divided by the through
                    thickness count to give the lead element height. The default value is 2 and the
                    maximum is 10. Only used when the advanced lead mesh setting is enabled.
        display_model: bool
            Whether to display the model after export.
        clear_FEA_database: bool
            Whether to clear FEA database before defining model.
        use_FEA_model_id: bool
            Whether to use FEA model ID.
        coordinate_units: str
            Units of the model coordinates to use when exporting a model.


        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.launcher import launch_sherlock
        >>> from ansys.sherlock.core.types.common_types import (
            Measurement,
        )
        >>> sherlock = launch_sherlock()
        >>> sherlock.model.export_FEA_model(
                project="Test Project",
                cca_name="Main Board",
                export_file="C:/Temp/export.wbjn",
                analysis="NaturalFreq",
                drill_hole_parameters=[
                    {
                        "drill_hole_modeling": "ENABLED",
                        "min_hole_diameter": Measurement(value=0.5, unit="mm"),
                        "max_edge_length": Measurement(value=1.0, unit="mm")
                    }
                ],
                detect_lead_modeling="ENABLED",
                lead_model_parameters=[
                    {
                        "lead_modeling": "ENABLED",
                        "lead_element_order": "First Order (Linear)",
                        "max_mesh_size": Measurement(value=0.5, unit="mm"),
                        "vertical_mesh_size": Measurement(value=0.1, unit="mm"),
                        "thicknessCount": 3,
                        "aspectRatio": 2
                    }
                ],
                display_model=True,
                clear_FEA_database=True,
                use_FEA_model_id=True,
                coordinate_units="mm"
            )
        """
        try:
            if not project:
                raise SherlockExportFEAModelError(message="Project name is invalid.")

            if not cca_name:
                raise SherlockExportFEAModelError(message="CCA name is invalid.")

            if not export_file:
                raise SherlockExportFEAModelError(message="Export file path is invalid.")

            if not os.path.exists(os.path.dirname(export_file)):
                raise SherlockExportFEAModelError(
                    message=f"Export file directory " f'"{export_file}" ' f"does not exist."
                )

            for param in drill_hole_parameters:
                min_hole_diameter = param.get("min_hole_diameter")
                if not isinstance(min_hole_diameter, Measurement):
                    raise SherlockExportFEAModelError(message="Minimum hole diameter is invalid.")

                max_edge_length = param.get("max_edge_length")
                if not isinstance(max_edge_length, Measurement):
                    raise SherlockExportFEAModelError(message="Maximum edge length is invalid.")

            for param in lead_model_parameters:
                max_mesh_size = param.get("max_mesh_size")
                if not isinstance(max_mesh_size, Measurement):
                    raise SherlockExportFEAModelError(message="Maximum mesh size is invalid.")

                vertical_mesh_size = param.get("vertical_mesh_size")
                if not isinstance(vertical_mesh_size, Measurement):
                    raise SherlockExportFEAModelError(message="Vertical mesh size is invalid.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            export_request = SherlockModelService_pb2.ExportFEAModelRequest()
            export_request.project = project
            export_request.ccaName = cca_name
            export_request.exportFile = export_file
            export_request.analysis = (
                SherlockModelService_pb2.ExportFEAModelRequest.ExportAnalysis.Value(analysis)
            )

            for param in drill_hole_parameters:
                export_request.drillHoleParam.drillHoleModeling = param.get("drill_hole_modeling")

                min_hole_diameter = param.get("min_hole_diameter")
                export_request.drillHoleParam.minHoleDiameter.value = min_hole_diameter.value
                export_request.drillHoleParam.minHoleDiameter.unit = min_hole_diameter.unit

                max_edge_length = param.get("max_edge_length")
                export_request.drillHoleParam.maxEdgeLength.value = max_edge_length.value
                export_request.drillHoleParam.maxEdgeLength.unit = max_edge_length.unit

            export_request.detectLeadModeling = detect_lead_modeling.upper()

            for param in lead_model_parameters:
                export_request.leadModelParam.leadModeling = param.get("lead_modeling").upper()

                export_request.leadModelParam.leadElemOrder = param.get("lead_element_order")

                max_mesh_size = param.get("max_mesh_size")
                export_request.leadModelParam.maxMeshSize.value = max_mesh_size.value
                export_request.leadModelParam.maxMeshSize.unit = max_mesh_size.unit

                vertical_mesh_size = param.get("vertical_mesh_size")
                export_request.leadModelParam.verticalMeshSize.value = vertical_mesh_size.value
                export_request.leadModelParam.verticalMeshSize.unit = vertical_mesh_size.unit

                thickness_count = param.get("thicknessCount", 3)
                export_request.leadModelParam.thicknessCount = thickness_count

                aspect_ratio = param.get("aspectRatio", 2)
                export_request.leadModelParam.aspectRatio = aspect_ratio

            export_request.displayModel = display_model
            export_request.clearFEADatabase = clear_FEA_database
            export_request.useFEAModelID = use_FEA_model_id
            export_request.coordinateUnits = coordinate_units

            return_code = self.stub.exportFEAModel(export_request)
            if return_code.value != 0:
                raise SherlockExportFEAModelError(message=return_code.message)

            return return_code.value

        except SherlockExportFEAModelError as e:
            LOG.error(str(e))
            raise e
